from collections import defaultdict

import distances as nm_dist


def make_distance_metrics(
    indel: bool | dict = False,
    discounted_levenshtein: bool | dict = False,
    levenshtein: bool | dict = False,
    jaro_winkler: bool | dict = False,
    tichy: bool | dict = False,
    cormodel_z: bool | dict = False,
    iterative_sub_string: bool | dict = False,
    baulieu_xiii: bool | dict = False,
    clement: bool | dict = False,
    dice_asymmetrici: bool | dict = False,
    kuhns_iii: bool | dict = False,
    overlap: bool | dict = False,
    pearson_ii: bool | dict = False,
    weighted_jaccard: bool | dict = False,
    warrens_iv: bool | dict = False,
    bag: bool | dict = False,
    rouge_l: bool | dict = False,
    token_distance: bool | dict = False,
    ratcliff_obershelp: bool | dict = False,
    ncd_bz2: bool | dict = False,
    fuzzy_wuzzy_partial_string: bool | dict = False,
    fuzzy_wuzzy_token_sort: bool | dict = False,
    fuzzy_wuzzy_token_set: bool | dict = False,
    editex: bool | dict = False,
    typo: bool | dict = False,
    lig_3: bool | dict = False,
    ssk: bool | dict = False,
    refined_soundex: bool | dict = False,
    double_metaphone: bool | dict = False,
) -> dict:
    r"""
    A function which returns a dict containing the distance metrics that should be
    used during the fuzzy string matching

    Levenshtein edit distance
        - Indel
        - Discounted Levenshtein
        - levenshtein
        - LIG3
        - Jaro-Winkler
    Block edit distances
        - Tichy
        - CormodeLZ
    Multi-set token-based distance
        - BaulieuXIII
        - Clement
        - DiceAsymmetricI
        - KuhnsIII
        - Overlap
        - PearsonII
        - WeightedJaccard
        - WarrensIV
        - Bag
        - RougeL
        - Token distance
    Subsequence distances
        - IterativeSubString
        - RatcliffObershelp
        - SSK
    Normalized compression distance
        - NCDbz2
    FuzzyWuzzy distances
        - FuzzyWuzzyPartialString
        - FuzzyWuzzyTokenSort
        - FuzzyWuzzyTokenSet
    Ponetic distances
        - RefinedSoundex
        - DoubleMetaphone
    Edit distances
        - Editex
        - Typo


    Parameters
    ----------
    indel: bool
        Boolean indicating whether the Indel method should be used during the
        fuzzy name matching. The indel method is equal to a regular levenshtein
        distance with a twice as high substitution weight. If a dictionary is provided,
        it is used as parameters for the indel distance metric.
        default=False
    discounted_levenshtein: bool
        Boolean indicating whether the DiscountedLevenshtein method should be used
        during the fuzzy name matching. Equal to the regular levenshtein distance,
        only errors later in the string are counted at a discounted rate. To limit
        the importance of for instance suffix differences. If a dictionary is provided,
        it is used as parameters for the discounted_levenshtein distance metric.
        default=False
    levenshtein: bool
        Boolean indicating whether the Levenshtein method should be used
        during the fuzzy name matching. If a dictionary is provided,
        it is used as parameters for the levenshtein distance metric.
        default=False
    jaro_winkler: bool
        Boolean indicating whether the JaroWinkler method should be used
        during the fuzzy name matching. If a dictionary is provided,
        it is used as parameters for the jaro winkler distance metric.
        default=False
    tichy: bool
        Boolean indicating whether the Tichy method should be used during the
        fuzzy name matching. This algorithm provides a shortest edit distance based
        on substring and add operations. If a dictionary is provided,
        it is used as parameters for the tichy distance metric.
        default=False
    cormodel_z: bool
        Boolean indicating whether the CormodeLZ method should be used during the
        fuzzy name matching. The CormodeLZ distance between strings x and y, is the
        minimum number of single characters or substrings of y or of the partially
        built string which are required to produce x from left to right. If a dictionary is provided,
        it is used as parameters for the cormodel_z distance metric.
        default=False
    iterative_sub_string: bool
        Boolean indicating whether the IterativeSubString method should be used
        during the fuzzy name matching. A method that counts the similarities
        between two strings substrings and subtracts the differences taking into
        account the winkler similarity between the string and the substring. If a dictionary is provided,
        it is used as parameters for the iterative_sub_string distance metric.
        default=False
    baulieu_xiii: bool
        Boolean indicating whether the BaulieuXIII method should be used during
        the fuzzy name matching. The Baulieu XIII distance between two strings is
        given by the following formula: (|X \ Y| + |Y \ X|) / (
        |X ∩ Y| + |X \ Y| + |Y \ X| + |X ∩ Y| ∙ (|X ∩ Y| - 4)^2). If a dictionary is provided,
        it is used as parameters for the baulieu_xiii distance metric.
        default=False
    clement: bool
        Boolean indicating whether the Clement method should be used during the
        fuzzy name matching. The Clement distance between two strings is given
        by the following formula: (|X ∩ Y|/|X|)*(1-|X|/|N|) + (|(N \ X) \ Y|/|N \ X|) *
        (1-|N \ X|/|N|). If a dictionary is provided,
        it is used as parameters for the clement distance metric.
        default=False
    dice_asymmetrici: bool
        Boolean indicating whether the DiceAsymmetricI method should be used during
        the fuzzy name matching. The Dice asymmetric similarity is given be |X ∩ Y|/|X|. If a dictionary is provided,
        it is used as parameters for the dice_asymmetrici distance metric.
        default=False
    kuhns_iii: bool
        Boolean indicating whether the KuhnsIII method should be used during the
        fuzzy name matching. If a dictionary is provided,
        it is used as parameters for the kuhns_iii distance metric.
        default=False
    overlap: bool
        Boolean indicating whether the Overlap method should be used during the
        fuzzy name matching. The overlap distance is given by: |X ∩ Y|/min(|X|,|Y|). If a dictionary is provided,
        it is used as parameters for the overlap distance metric.
        default=True
    pearson_ii: bool
        Boolean indicating whether the PearsonII method should be used during the
        fuzzy name matching. This algorithm is based on the Phi coefficient or the
        mean square contingency. If a dictionary is provided,
        it is used as parameters for the pearson_ii distance metric.
        default=False
    weighted_jaccard: bool
        Boolean indicating whether the WeightedJaccard method should be used during
        the fuzzy name matching. This is the Jaccard distance only using a wheighing
        for the differences of 3. If a dictionary is provided,
        it is used as parameters for the weighted_jaccard distance metric.
        default=True
    warrens_iv: bool
        Boolean indicating whether the WarrensIV method should be used during the
        fuzzy name matching. If a dictionary is provided,
        it is used as parameters for the warrens_iv distance metric.
        default=False
    bag: bool
        Boolean indicating whether the Bag method should be used during the fuzzy
            name matching. Is a simplification of the regular edit distance by using
            a similarity tree structure. If a dictionary is provided,
        it is used as parameters for the bag distance metric.
        default=False
    rouge_l: bool
        Boolean indicating whether the ROUGE-L method should be used during the
        fuzzy name matching. The ROGUE-L method is a measure that counts the longest
        substring between to strings. If a dictionary is provided,
        it is used as parameters for the rouge_l distance metric.
        default=False
    token_distance: bool
        Boolean indicating whether the a token distance method should be used during the
        fuzzy name matching. If a dictionary is provided, it is used as parameters for
        the tokenised distance metrics.
        default=False
    ratcliff_obershelp: bool
        Boolean indicating whether the RatcliffObershelp method should be used
        during the fuzzy name matching. This method finds the longest common substring
        and evaluates the longest common substrings to the right and the left of the
        original longest common substring. If a dictionary is provided,
        it is used as parameters for the ratcliff_obershelp distance metric.
        default=True
    ncd_bz2: bool
        Boolean indicating whether the NCDbz2 method should be used during the
        fuzzy name matching. Applies the Burrows-Wheeler transform to the strings and
        subsequently returns the normalised compression distance. If a dictionary is provided,
        it is used as parameters for the ncd_bz2 distance metric.
        default=False
    fuzzy_wuzzy_partial_string: bool
        Boolean indicating whether the FuzzyWuzzyPartialString method should be used
        during the fuzzy name matching. This methods takes the length of the longest
        common substring and divides it over the minimum of the length of each of
        the two strings. If a dictionary is provided,
        it is used as parameters for the fuzzy_wuzzy_partial_string distance metric.
        default=False
    fuzzy_wuzzy_token_sort: bool
        Boolean indicating whether the FuzzyWuzzyTokenSort method should be used
        during the fuzzy name matching. This tokenizes the words in the string
        and sorts them, subsequently a hamming distance is calculated. If a dictionary is provided,
        it is used as parameters for the fuzzy_wuzzy_token_sort distance metric.
        default=True
    fuzzy_wuzzy_token_set: bool
        Boolean indicating whether the FuzzyWuzzyTokenSet method should be used
        during the fuzzy name matching. This method tokenizes the strings and
        find the largest intersection of the two substrings and divides it over
        the length of the shortest string. If a dictionary is provided,
        it is used as parameters for the fuzzy_wuzzy_token_set distance metric.
        default=False
    editex: bool
        Boolean indicating whether the Editex method should be used during the
        fuzzy name matching. If a dictionary is provided,
        it is used as parameters for the editex distance metric.
        default=True
    typo: bool
        Boolean indicating whether the Typo method should be used during the
        fuzzy name matching. The typo distance is calculated based on the distance
        on a keyboard between edits. If a dictionary is provided,
        it is used as parameters for the typo distance metric.
        default=False
    lig_3: bool
        Boolean indicating whether the LIG3 method should be used during the fuzzy
        name matching. If a dictionary is provided,
        it is used as parameters for the lig_3 distance metric.
        default=False
    ssk: bool
        Boolean indicating whether the SSK method should be used during the fuzzy
        name matching. The ssk algorithm looks at the string kernel generated by all
        the possible different subsequences present between the two strings. If a dictionary is provided,
        it is used as parameters for the ssk distance metric.
        default=False
    refined_soundex: bool
        Boolean indicating whether the string should be represented by the RefinedSoundex
        phonetix translation and the Levensthein distance of the translated strings should
        be included in the fuzzy matching process. If a dictionary is provided,
        it is used as parameters for the refined_soundex distance metric.
        default=False
    double_metaphone: bool
        Boolean indicating whether the string should be represented by the DoubleMetaphone
        phonetix translation and the Levensthein distance of the translated strings should
        be included in the fuzzy matching process. If a dictionary is provided,
        it is used as parameters for the Double Metaphone distance metric.
        default=False

    """
    distance_metrics = defaultdict(list)
    if indel:
        if isinstance(indel, dict):
            distance_metrics["Levenshtein"].append(nm_dist.Indel(**indel))
        else:
            distance_metrics["Levenshtein"].append(nm_dist.Indel())
    if discounted_levenshtein:
        if isinstance(discounted_levenshtein, dict):
            distance_metrics["Levenshtein"].append(
                nm_dist.DiscountedLevenshtein(**discounted_levenshtein)
            )
        else:
            distance_metrics["Levenshtein"].append(nm_dist.DiscountedLevenshtein())
    if levenshtein:
        if isinstance(levenshtein, dict):
            distance_metrics["Levenshtein"].append(nm_dist.Levenshtein(**levenshtein))
        else:
            distance_metrics["Levenshtein"].append(nm_dist.Levenshtein())
    if jaro_winkler:
        if isinstance(jaro_winkler, dict):
            distance_metrics["Levenshtein"].append(nm_dist.JaroWinkler(**jaro_winkler))
        else:
            distance_metrics["Levenshtein"].append(nm_dist.JaroWinkler())
    if cormodel_z:
        if isinstance(cormodel_z, dict):
            distance_metrics["block"].append(nm_dist.CormodeLZ(**cormodel_z))
        else:
            distance_metrics["block"].append(nm_dist.CormodeLZ())
    if tichy:
        if isinstance(tichy, dict):
            distance_metrics["block"].append(nm_dist.Tichy(**tichy))
        else:
            distance_metrics["block"].append(nm_dist.Tichy())
    if iterative_sub_string:
        if isinstance(iterative_sub_string, dict):
            distance_metrics["Subsequence"].append(
                nm_dist.IterativeSubString(**iterative_sub_string)
            )
        else:
            distance_metrics["Subsequence"].append(nm_dist.IterativeSubString())
    if baulieu_xiii:
        if isinstance(baulieu_xiii, dict):
            distance_metrics["multiset"].append(nm_dist.BaulieuXIII(**baulieu_xiii))
        else:
            distance_metrics["multiset"].append(nm_dist.BaulieuXIII())
    if clement:
        if isinstance(clement, dict):
            distance_metrics["multiset"].append(nm_dist.Clement(**clement))
        else:
            distance_metrics["multiset"].append(nm_dist.Clement())
    if dice_asymmetrici:
        if isinstance(dice_asymmetrici, dict):
            distance_metrics["multiset"].append(
                nm_dist.DiceAsymmetricI(**dice_asymmetrici)
            )
        else:
            distance_metrics["multiset"].append(nm_dist.DiceAsymmetricI())
    if kuhns_iii:
        if isinstance(kuhns_iii, dict):
            distance_metrics["multiset"].append(nm_dist.KuhnsIII(**kuhns_iii))
        else:
            distance_metrics["multiset"].append(nm_dist.KuhnsIII())
    if overlap:
        if isinstance(overlap, dict):
            distance_metrics["multiset"].append(nm_dist.Overlap(**overlap))
        else:
            distance_metrics["multiset"].append(nm_dist.Overlap())
    if pearson_ii:
        if isinstance(pearson_ii, dict):
            distance_metrics["multiset"].append(nm_dist.PearsonII(**pearson_ii))
        else:
            distance_metrics["multiset"].append(nm_dist.PearsonII())
    if weighted_jaccard:
        if isinstance(weighted_jaccard, dict):
            distance_metrics["multiset"].append(
                nm_dist.WeightedJaccard(**weighted_jaccard)
            )
        else:
            distance_metrics["multiset"].append(nm_dist.WeightedJaccard())
    if warrens_iv:
        if isinstance(warrens_iv, dict):
            distance_metrics["multiset"].append(nm_dist.WarrensIV(**warrens_iv))
        else:
            distance_metrics["multiset"].append(nm_dist.WarrensIV())
    if bag:
        if isinstance(bag, dict):
            distance_metrics["multiset"].append(nm_dist.Bag(**bag))
        else:
            distance_metrics["multiset"].append(nm_dist.Bag())
    if rouge_l:
        if isinstance(rouge_l, dict):
            distance_metrics["multiset"].append(nm_dist.RougeL(**rouge_l))
        else:
            distance_metrics["multiset"].append(nm_dist.RougeL())
    if token_distance:
        if isinstance(token_distance, dict):
            distance_metrics["multiset"].append(
                nm_dist._TokenDistance(**token_distance)
            )
        else:
            distance_metrics["multiset"].append(nm_dist._TokenDistance())
    if ratcliff_obershelp:
        if isinstance(ratcliff_obershelp, dict):
            distance_metrics["Subsequence"].append(
                nm_dist.RatcliffObershelp(**ratcliff_obershelp)
            )
        else:
            distance_metrics["Subsequence"].append(nm_dist.RatcliffObershelp())
    if ncd_bz2:
        if isinstance(ncd_bz2, dict):
            distance_metrics["compression"].append(nm_dist.NCDbz2(**ncd_bz2))
        else:
            distance_metrics["compression"].append(nm_dist.NCDbz2())
    if fuzzy_wuzzy_partial_string:
        if isinstance(fuzzy_wuzzy_partial_string, dict):
            distance_metrics["fuzzy"].append(
                nm_dist.FuzzyWuzzyPartialString(**fuzzy_wuzzy_partial_string)
            )
        else:
            distance_metrics["fuzzy"].append(nm_dist.FuzzyWuzzyPartialString())
    if fuzzy_wuzzy_token_sort:
        if isinstance(fuzzy_wuzzy_token_sort, dict):
            distance_metrics["fuzzy"].append(
                nm_dist.FuzzyWuzzyTokenSort(**fuzzy_wuzzy_token_sort)
            )
        else:
            distance_metrics["fuzzy"].append(nm_dist.FuzzyWuzzyTokenSort())
    if fuzzy_wuzzy_token_set:
        if isinstance(fuzzy_wuzzy_token_set, dict):
            distance_metrics["fuzzy"].append(
                nm_dist.FuzzyWuzzyTokenSet(**fuzzy_wuzzy_token_set)
            )
        else:
            distance_metrics["fuzzy"].append(nm_dist.FuzzyWuzzyTokenSet())
    if editex:
        if isinstance(editex, dict):
            distance_metrics["edit"].append(nm_dist.Editex(**editex))
        else:
            distance_metrics["edit"].append(nm_dist.Editex())
    if typo:
        if isinstance(typo, dict):
            distance_metrics["edit"].append(nm_dist.Typo(**typo))
        else:
            distance_metrics["edit"].append(nm_dist.Typo())
    if lig_3:
        if isinstance(lig_3, dict):
            distance_metrics["Levenshtein"].append(nm_dist.LIG3(**lig_3))
        else:
            distance_metrics["Levenshtein"].append(nm_dist.LIG3())
    if ssk:
        if isinstance(ssk, dict):
            distance_metrics["Subsequence"].append(nm_dist.SSK(**ssk))
        else:
            distance_metrics["Subsequence"].append(nm_dist.SSK())
    if refined_soundex:
        if not isinstance(refined_soundex, dict):
            refined_soundex = {"max_length": 30}
        distance_metrics["phonetic"].append(
            nm_dist.PhoneticDistance(
                transforms=nm_dist.RefinedSoundex(**refined_soundex),
                metric=nm_dist.Levenshtein(),
                encode_alpha=True,
            )
        )
    if double_metaphone:
        if not isinstance(double_metaphone, dict):
            double_metaphone = {"max_length": 30}
        distance_metrics["phonetic"].append(
            nm_dist.PhoneticDistance(
                transforms=nm_dist.DoubleMetaphone(**double_metaphone),
                metric=nm_dist.Levenshtein(),
                encode_alpha=True,
            )
        )

    return distance_metrics



import importlib.resources as importlib_resource
import os
import pickle
import re
import warnings
from functools import reduce
from itertools import compress
from operator import iconcat
from re import escape, sub
from typing import List, Optional, Tuple, Union
from unicodedata import category, normalize

import numpy as np
import pandas as pd
from sklearn.feature_extraction.text import TfidfVectorizer
from tqdm import tqdm

from name_matching.data.transliterations import TRANSLITERATION_MAP
from name_matching.distance_metrics import make_distance_metrics
from name_matching.sparse_cosine import sparse_cosine_top_n


class NameMatcher:
    """
    Name matching using character n-gram cosine similarity followed by fuzzy matching.

    This class first vectorizes names using character n-grams (via TF-IDF and cosine
    similarity), selects the top N candidates, and then applies multiple fuzzy matching
    distance metrics to pick the best match or top-K matches.

    Parameters
    ----------
    ngrams : tuple of int, default=(2, 3)
        Character n-gram lengths used for cosine similarity.
    top_n : int, default=50
        Number of top candidates to return from the cosine step.
    low_memory : bool, default=False
        If True, uses a low-memory approach for sparse cosine similarity.
    number_of_rows : int, default=5000
        Batch size for low-memory sparse cosine similarity. Ignored if low_memory is True.
    number_of_matches : int, default=1
        Number of fuzzy-matched alternatives to return.
    lowercase : bool, default=True
        If True, converts text to lowercase during preprocessing.
    non_word_characters : Optional[bool], default=True
        If True, strips non-word characters (excluding & and #) during preprocessing.
    remove_ascii : bool, default=True
        If True, transliterates to ASCII (dropping accents) during preprocessing.
    punctuations : Optional[bool], default=None
        Deprecated alias for non_word_characters.
    legal_suffixes : bool, default=False
        If True, post-processing will ignore common company legal suffixes in scoring.
    preprocess_legal : bool, default=False
        If True, strips or abbreviates legal suffixes/prefixes during preprocessing.
    delete_legal : bool, default=False
        If True, deletes legal suffixes/prefixes instead of abbreviating them.
    make_abbreviations : bool, default=True
        If True, replaces common words with their abbreviations during preprocessing.
    common_words : Union[bool, list], default=False
        If True, will post-process to down-weight the most common words. If a list is
        provided, those specific words will be down-weighted.
    cut_off_no_scoring_words : float, default=0.01
        Threshold (fraction of max frequency) above which a word is considered too common.
    preprocess_split : bool, default=False
        If True, performs an additional “split” variant of preprocessing for searching.
    begin_end_legal_pre_suffix : bool, default=True
        If True, only abbreviate legal terms at the beginning or end of names.
    verbose : bool, default=True
        If True, prints progress via tqdm.
    distance_metrics : list of str, default=[
        "overlap", "weighted_jaccard", "ratcliff_obershelp",
        "fuzzy_wuzzy_token_sort", "editex"]
        List of distance metric names to use in the fuzzy-matching step.
    row_numbers : bool, default=False
        If True, returns original DataFrame index values in the match results.
    return_algorithms_score : bool, default=False
        If True, return the full per-algorithm score matrix instead of just combined scores.
    save_intermediate_results : bool, default=False
        If True, saves intermediate pickle files for matching_data, to_be_matched, possible_matches.
    load_intermediate_results : bool, default=False
        If True, attempts to load intermediate pickle files before recomputing.
    intermediate_results_name : dict of str to str, default={
        "matching_data": "df_matching_data_name",
        "to_be_matched": "to_be_matched_name",
        "possible_matches": "possible_matches_name"
    }
        Filenames (without “.pkl”) for saving/loading intermediate results.

    Raises
    ------
    TypeError
        If `common_words` is not a bool or iterable of strings.
    """

    def __init__(
        self,
        ngrams: tuple = (2, 3),
        top_n: int = 50,
        low_memory: bool = False,
        number_of_rows: int = 5000,
        number_of_matches: int = 1,
        lowercase: bool = True,
        non_word_characters: bool = None,
        remove_ascii: bool = True,
        punctuations: bool = None,
        legal_suffixes: bool = False,
        preprocess_legal: bool = False,
        delete_legal: bool = False,
        make_abbreviations: bool = True,
        common_words: Union[bool, list] = False,
        cut_off_no_scoring_words: float = 0.01,
        preprocess_split: bool = False,
        begin_end_legal_pre_suffix: bool = True,
        verbose: bool = True,
        distance_metrics: Union[list, tuple] = [
            "overlap",
            "weighted_jaccard",
            "ratcliff_obershelp",
            "fuzzy_wuzzy_token_sort",
            "editex",
        ],
        row_numbers: bool = False,
        return_algorithms_score: bool = False,
        save_intermediate_results: bool = False,
        load_intermediate_results: bool = False,
        intermediate_results_name: dict[str, str] = {
            "matching_data": "df_matching_data_name",
            "to_be_matched": "to_be_matched_name",
            "possible_matches": "possible_matches_name",
        },
    ):

        self._possible_matches = None
        self._preprocessed = False
        self._df_matching_data = pd.DataFrame()

        self._number_of_rows = number_of_rows
        self._low_memory = low_memory
        self._save = save_intermediate_results
        self._load = load_intermediate_results
        self._intermediate_results_name = intermediate_results_name

        self._column = ""
        self._column_matching = ""

        self._verbose = verbose
        self._number_of_matches = number_of_matches
        self._top_n = top_n
        self._return_algorithms_score = return_algorithms_score

        self._preprocess_lowercase = lowercase
        self._preprocess_non_word_characters = self._process_punctuations_depricated(
            non_word_characters, punctuations
        )
        self._preprocess_ascii = remove_ascii
        self._preprocess_abbreviations = make_abbreviations
        self._preprocess_legal_suffixes = preprocess_legal
        self._postprocess_company_legal_id = legal_suffixes
        self._delete_legal = delete_legal

        if isinstance(common_words, bool):
            self._postprocess_common_words = common_words
            self._word_set = set()
        elif isinstance(common_words, (list, tuple, set)):
            self._postprocess_common_words = False
            self._word_set = set(common_words)
        else:
            raise TypeError("Please provide common_words as a list or a bool")

        self._preprocess_split = preprocess_split
        self._cut_off = cut_off_no_scoring_words

        if self._postprocess_company_legal_id:
            self._word_set = self._make_no_scoring_words(
                "legal", self._word_set, self._cut_off
            )

        self._original_indexes = not row_numbers
        self._original_index = None
        self._begin_end_legal_pre_suffix = begin_end_legal_pre_suffix

        self.set_distance_metrics(distance_metrics)  # type: ignore

        self._vec = TfidfVectorizer(
            lowercase=False, analyzer="char", ngram_range=(ngrams)
        )
        self._n_grams_matching = None

    def _process_punctuations_depricated(
        self, non_word_characters: bool, punctuations: bool
    ) -> bool:
        """
        Handle backward compatibility between `punctuations` and `non_word_characters`.

        Parameters
        ----------
        non_word_characters : bool
            New parameter indicating whether to strip non-word characters.
        punctuations : bool
            Deprecated parameter (alias of `non_word_characters`).

        Returns
        -------
        bool
            Final value to use for stripping non-word characters.

        Warnings
        --------
        Warns if both parameters are provided but disagree.
        """

        if non_word_characters is None:
            if punctuations is None:
                return True
            else:
                warnings.warn(
                    "The punctuations parameter has been replaced with the non_word_characters "
                    + "parameter, please use the non_word_characters parameter going forward. The"
                    + "punctuations parameter will be depricated in the future."
                )
                return punctuations
        else:
            if (punctuations is not None) & (punctuations != non_word_characters):
                warnings.warn(
                    "non_word_characters is the new name of the punctuations parameter. "
                    + "These parameters are now not equal and the non_word_characters is used."
                    + "The punctuations parameter will be depricated in the future."
                )
            return non_word_characters

    def _generate_combinations(
        self,
        list_a: List,
        list_b: List,
        ind: int = 0,
        result: Optional[List] = None,
    ) -> None:
        """
        Recursively build all possible element-wise choices between two lists.

        At each index `i`, you may choose `list_a[i]` or `list_b[i]`.

        Parameters
        ----------
        list_a : list
            First choice list.
        list_b : list
            Second choice list.
        ind : int, default=0
            Current index in recursion.
        result : list, optional
            Sequence built so far.

        Returns
        -------
        None
            Results are appended to `self._temp`.
        """
        if result is None:
            result = []

        if ind == len(list_a):
            self._temp.append(result)
            return

        self._generate_combinations(list_a, list_b, ind + 1, result + [list_a[ind]])
        self._generate_combinations(list_b, list_a, ind + 1, result + [list_b[ind]])

    def _replace_substring(
        self,
        name: str,
        abbreviations: List[str],
        long_names: List[str],
        begin_end: bool = True,
        delete_names: bool = False,
    ) -> str:
        """
        Replace or delete substrings in `name` according to provided maps.

        Parameters
        ----------
        name : str
            Original string.
        abbreviations : list of str
            Short forms to insert.
        long_names : list of str
            Corresponding full forms to replace.
        begin_end : bool, default=True
            If True, only replace when the full form is at the start or end.
        delete_names : bool, default=False
            If True, delete the matched substring entirely instead of abbreviating.

        Returns
        -------
        str
            Modified string.
        """
        if begin_end:
            for abbreviation, long_name in zip(abbreviations, long_names):
                if name.startswith(long_name) | name.endswith(long_name):
                    if delete_names:
                        name = re.sub(rf"\b{long_name}$", "", name)
                        name = re.sub(rf"^{long_name}\b", "", name)
                    else:
                        name = re.sub(rf"\b{long_name}$", abbreviation, name)
                        name = re.sub(rf"^{long_name}\b", abbreviation, name)
        else:
            for abbreviation, long_name in zip(abbreviations, long_names):
                if long_name in name:
                    if delete_names:
                        name = re.sub(rf"\b{long_name}\b", "", name)
                    else:
                        name = re.sub(rf"\b{long_name}\b", abbreviation, name)

        return name

    def _replace_common_strings(
        self, data: pd.DataFrame, column_name: str
    ) -> pd.DataFrame:
        """
        Abbreviate common words in a DataFrame column using an external CSV map.

        Parameters
        ----------
        data : pd.DataFrame
            DataFrame to modify.
        column_name : str
            Column on which to apply replacements.

        Returns
        -------
        pd.DataFrame
            Modified DataFrame.
        """
        with importlib_resource.as_file(
            importlib_resource.files("name_matching.data").joinpath("common_words.csv")
        ) as path:
            common_words = pd.read_csv(path)
        data.loc[:, column_name] = data.loc[:, column_name].apply(
            lambda x: self._replace_substring(
                x,
                common_words["short_form"].tolist(),
                common_words["word"].tolist(),
                begin_end=False,
            )
        )

        return data

    def _replace_legal_pre_suffixes_with_abbreviations(
        self, data: pd.DataFrame, column_name: str
    ) -> pd.DataFrame:
        """
        Abbreviate or delete legal prefixes/suffixes in a DataFrame column.
        Reads `legal_names.csv` to get mappings.

        Parameters
        ----------
        data : pd.DataFrame
            DataFrame containing the column.
        column_name : str
            Name of the column to process.

        Returns
        -------
        pd.DataFrame
            DataFrame with legal terms replaced.
        """
        abbreviations = []
        possible_names = []
        with importlib_resource.as_file(
            importlib_resource.files("name_matching.data").joinpath("legal_names.csv")
        ) as path:
            legal_words = pd.read_csv(path)

        for _, legal_word in legal_words.iterrows():
            abbr = re.split(r"[. /]", legal_word["abbreviation"].strip().lower())
            abbr = list(filter(None, abbr))
            lgl = legal_word["full_name"].lower().strip().split(" ")

            if len(abbr) == len(lgl):
                self._temp = []
                self._generate_combinations(abbr, lgl)
            elif len(abbr) < len(lgl):
                new_lgl = self._combine_legal_words(abbr, lgl)
                if len(new_lgl) == len(abbr):
                    self._temp = []
                    self._generate_combinations(abbr, new_lgl)
                else:
                    self._temp = [legal_word["full_name"]]
            else:
                self._temp = [legal_word["full_name"]]

            self._temp.append("".join(abbr))

            for option in self._temp:
                abbreviations.append(legal_word["abbreviation"].lower())
                if self._preprocess_non_word_characters:
                    possible_names.append(
                        option.strip()
                        if isinstance(option, str)
                        else " ".join(option).strip()
                    )
                else:
                    if isinstance(option, str):
                        possible_names.append(option.strip())
                    else:
                        possible_names.append(" ".join(option).strip())
                        possible_names.append(".".join(option).strip() + ".")
                        abbreviations.append(legal_word["abbreviation"].lower())

        if self._delete_legal:
            possible_names.sort(key=len, reverse=True)

        data[column_name] = data.apply(
            lambda x: self._replace_substring(
                x[column_name],
                abbreviations,
                possible_names,
                begin_end=self._begin_end_legal_pre_suffix,
                delete_names=self._delete_legal,
            ),
            axis=1,
        )

        return data

    def _combine_legal_words(self, abbr: List[str], lgl: List[str]) -> List[str]:
        """
        Merge parts of a full legal name so they align with its abbreviation parts.

        Parameters
        ----------
        abbr : list of str
            Segments of the abbreviation.
        lgl : list of str
            Segments of the full legal name.

        Returns
        -------
        list of str
            Re-grouped segments of the legal name.
        """
        ind = 0
        new_lgl = []
        combined_name = ""
        for letter in abbr:
            while ind < len(lgl) and not lgl[ind].startswith(letter):
                combined_name += " " + lgl[ind]
                ind += 1
            if ind < len(lgl) and lgl[ind].startswith(letter):
                if combined_name:
                    new_lgl.append(combined_name.strip())
                combined_name = lgl[ind]
                ind += 1
        if combined_name:
            new_lgl.append(combined_name.strip())
        return new_lgl

    def set_distance_metrics(self, metrics: List) -> None:
        """
        A method to set which of the distance metrics should be employed during the
        fuzzy matching. For very short explanations of most of the name matching
        algorithms please see the make_distance_metrics function in distance_matrics.py

        Parameters
        ----------
        metrics: List
            The list with the distance metrics to be used during the name matching. The
            distance metrics can be chosen from the list below:
                indel
                discounted_levenshtein
                tichy
                cormodeL_z
                iterative_sub_string
                baulieu_xiii
                clement
                dice_asymmetricI
                kuhns_iii
                overlap
                pearson_ii
                weighted_jaccard
                warrens_iv
                bag
                rouge_l
                q_grams
                ratcliff_obershelp
                ncd_bz2
                fuzzy_wuzzy_partial_string
                fuzzy_wuzzy_token_sort
                fuzzy_wuzzy_token_set
                editex
                typo
                lig_3
                ssk
                refined_soundex
                double_metaphone
        """

        input_metrics = {str(metric).lower(): True for metric in metrics}
        try:
            self._distance_metrics = make_distance_metrics(**input_metrics)
        except TypeError:
            raise TypeError(
                "Not all of the supplied distance metrics are available. Please check the"
                + "list of options in the make_distance_metrics function and adjust"
                + " your list accordingly"
            )
        self._num_distance_metrics = sum(
            [len(x) for x in self._distance_metrics.values()]
        )

    def _select_top_words(
        self, word: str, word_counts: pd.Series, occurrence_count: int
    ) -> str:
        """Remove the top words from the string word based on an occurrence_count threshold

        Parameters
        ----------
        word: str
            the string from which the words should be removed
        word_counts: pd.Series
            the words which should be removed with their counts as result from a value_counts
        occurrence_count: int
            the multiplication factor of the minimum occurrences below which to select

        Returns
        -------
        str
           The string word with the words with a too high word_counts removed
        """
        compressed_list = list(
            compress(
                word,
                (word_counts[word] < occurrence_count * word_counts[word].min()).values,
            )
        )

        return " ".join(compressed_list)

    def _preprocess_reduce(
        self, to_be_matched: pd.DataFrame, occurrence_count: int = 3
    ) -> pd.DataFrame:
        """Preprocesses and copies the data to obtain the data with reduced strings. The
        strings have all words removed which appear more than 3x as often as the least
        common word in the string and returns an adjusted copy of the input

        Parameters
        ----------
        to_be_matched: pd.DataFrame
            A dataframe from which the most common words should be removed
        occurrence_count: int
            The number of occurrence a word can occur more then the least common word in
            the string for which it will still be included in the process
            default=3

        Returns
        -------
        pd.DataFrame
            A dataframe that will contain the reduced strings
        """
        individual_words = (
            to_be_matched[self._column_matching].str.split(expand=True).stack()
        )
        word_counts = individual_words.value_counts()
        to_be_matched_new = to_be_matched.copy()
        name = to_be_matched[self._column_matching].str.split()
        to_be_matched_new[self._column_matching] = name.apply(
            lambda word: self._select_top_words(word, word_counts, occurrence_count)
        )

        return to_be_matched_new

    def load_and_process_master_data(
        self,
        column: str,
        df_matching_data: pd.DataFrame,
        start_processing: bool = True,
        transform: bool = True,
    ) -> None:
        """Load the matching data into the NameMatcher and start the preprocessing.

        Parameters
        ----------
        column : string
            The column name of the dataframe which should be used for the matching
        df_matching_data: pd.DataFrame
            The dataframe which is used to match the data to.
        start_processing : bool
            A boolean indicating whether to start the preprocessing step after
            loading the matching data. If transform is True the data will still be
            transformed and the preprocessing will be marked as completed.
            default: True
        transform : bool
            A boolean indicating whether or not the data should be transformed after
            the vectoriser is initialised
            default: True
        """
        self._column = column
        self._df_matching_data = df_matching_data
        self._original_index = df_matching_data.index
        if start_processing:
            self._process_matching_data(transform)
        elif transform:
            self._vectorise_data(transform)
            self._preprocessed = True

    def _process_matching_data(self, transform: bool = True) -> None:
        """Function to process the matching data. First the matching data is preprocessed
        and assigned to a variable within the NameMatcher. Next the data is used to
        initialise the TfidfVectorizer.

        Parameters
        ----------
        transform : bool
            A boolean indicating whether or not the data should be transformed after the
            vectoriser is initialised
            default: True
        """
        if self._load and os.path.exists(
            f"{self._intermediate_results_name['matching_data']}.pkl"
        ):
            with open(
                f"{self._intermediate_results_name['matching_data']}.pkl", "rb"
            ) as file:
                self._n_grams_matching = pickle.load(file).fillna("na")
        else:
            self._df_matching_data = self.preprocess(
                self._df_matching_data, self._column
            )

        if self._save:
            with open(
                f"{self._intermediate_results_name['matching_data']}.pkl", "wb"
            ) as file:
                pickle.dump(self._df_matching_data, file)

        if self._postprocess_common_words:
            self._word_set = self._make_no_scoring_words(
                "common", self._word_set, self._cut_off
            )

        self._vectorise_data(transform)
        self._preprocessed = True

    def match_names(
        self, to_be_matched: Union[pd.Series, pd.DataFrame], column_matching: str
    ) -> Union[pd.Series, pd.DataFrame] | Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Match input names against the preprocessed master data.

        This will:
          1. Preprocess the new names.
          2. Compute cosine-similarity top-N candidates.
          3. Apply fuzzy matching to those candidates.

        Parameters
        ----------
        to_be_matched : pandas Series or DataFrame
            New names to match.
        column_matching : str
            Column name in `to_be_matched` containing the names.

        Returns
        -------
        pandas Series or DataFrame
            If `return_algorithms_score=False` and `number_of_matches=1`, returns a
            DataFrame containing:
              - original_name
              - match_name
              - score
              - match_index

            If `return_algorithms_score=False` and `number_of_matches>1`, returns a
            DataFrame with columns for each alternative.

            If `return_algorithms_score=True`, returns a tuple (
            DataFrame_of_scores, DataFrame_of_matched_names ).
        """
        if self._column == "":
            raise ValueError(
                "Please first load the master data via the method: "
                + "load_and_process_master_data"
            )
        if self._verbose:
            tqdm.pandas()
            tqdm.write("preprocessing...\n")
        self._column_matching = column_matching

        is_dataframe = True
        if isinstance(to_be_matched, pd.Series):
            is_dataframe = False
            to_be_matched = pd.DataFrame(
                [to_be_matched.values], columns=to_be_matched.index.to_list()
            )

        if self._load and os.path.exists(
            f"{self._intermediate_results_name['to_be_matched']}.pkl"
        ):
            with open(
                f"{self._intermediate_results_name['to_be_matched']}.pkl", "rb"
            ) as file:
                to_be_matched = pickle.load(file).fillna("na")
        else:
            to_be_matched = self.preprocess(to_be_matched, self._column_matching)

        if self._save:
            with open(
                f"{self._intermediate_results_name['to_be_matched']}.pkl", "wb"
            ) as file:
                pickle.dump(to_be_matched, file)

        if self._load and os.path.exists(
            f"{self._intermediate_results_name['possible_matches']}.pkl"
        ):
            with open(
                f"{self._intermediate_results_name['possible_matches']}.pkl", "rb"
            ) as file:
                self._possible_matches = pickle.load(file)

            if not self._preprocessed:
                self._process_matching_data(False)

        else:
            if not self._preprocessed:
                self._process_matching_data()

            to_be_matched = self.preprocess(to_be_matched, self._column_matching)  # type: ignore
            if self._verbose:
                tqdm.write("preprocessing complete \n searching for matches...\n")
            self._possible_matches = self._search_for_possible_matches(to_be_matched)  # type: ignore

        if self._save:
            with open(
                f"{self._intermediate_results_name['possible_matches']}.pkl", "wb"
            ) as file:
                pickle.dump(self._possible_matches, file)

        if self._preprocess_split:
            self._possible_matches = np.hstack(
                (
                    self._search_for_possible_matches(
                        self._preprocess_reduce(to_be_matched)  # type: ignore
                    ),
                    self._possible_matches,
                )
            )

        if self._verbose:
            tqdm.write("possible matches found   \n fuzzy matching...\n")
            data_matches = to_be_matched.progress_apply(
                lambda x: self.fuzzy_matches(
                    self._possible_matches[to_be_matched.index.get_loc(x.name), :], x  # type: ignore
                ),
                axis=1,
            )  # type: ignore
        else:
            data_matches = to_be_matched.apply(
                lambda x: self.fuzzy_matches(
                    self._possible_matches[to_be_matched.index.get_loc(x.name), :], x  # type: ignore
                ),
                axis=1,
            )
        if self._return_algorithms_score:
            return data_matches, self._df_matching_data.iloc[
                self._possible_matches.flatten(), :
            ][
                self._column
            ].values.reshape(  # type: ignore
                (-1, self._top_n)
            )

        if self._number_of_matches == 1:
            data_matches = data_matches.rename(
                columns={
                    "match_name_0": "match_name",
                    "score_0": "score",
                    "match_index_0": "match_index",
                }
            )
        if is_dataframe and self._original_indexes:
            for col in data_matches.columns[
                data_matches.columns.str.contains("match_index")
            ]:
                data_matches[col] = self._original_index[  # type: ignore
                    data_matches[col].astype(int).fillna(0)
                ]  # type: ignore

        if self._verbose:
            tqdm.write("done")

        return data_matches

    def fuzzy_matches(
        self, possible_matches: np.array, to_be_matched: pd.Series  # type: ignore
    ) -> pd.Series:
        """A method which performs the fuzzy matching between the data in the
        to_be_matched series as well as the indicated indexes of the matching_data points
        which are possible matching candidates.

        Parameters
        ----------
        possible_matches : np.array
            An array containing the indexes of the matching data with potential matches
        to_be_matched : pd.Series
            The data which should be matched

        Returns
        -------
        pd.Series
            A series containing the match index from the matching_data dataframe. the name
            in the to_be_matched data, the name to which the datapoint was matched and a
            score between 0 (no match) and 100(perfect match) to indicate the quality of
            the matches.
        """
        if len(possible_matches.shape) > 1:
            possible_matches = possible_matches[0]

        indexes = np.array(
            [
                [f"match_name_{num}", f"score_{num}", f"match_index_{num}"]
                for num in range(self._number_of_matches)
            ]
        ).flatten()
        match = pd.Series(index=np.append("original_name", indexes), dtype=object)
        match["original_name"] = to_be_matched[self._column_matching]
        list_possible_matches = self._df_matching_data.iloc[
            possible_matches.flatten(), :
        ][self._column].values

        match_score = self._score_matches(
            to_be_matched[self._column_matching], list_possible_matches
        )
        if self._return_algorithms_score:
            return match_score
        ind = self._rate_matches(match_score)

        for num, col_num in enumerate(ind):
            match[f"match_name_{num}"] = list_possible_matches[col_num]
            match[f"match_index_{num}"] = possible_matches[col_num]

        match = self._adjust_scores(match_score[ind, :], match)

        if len(self._word_set):
            match = self.postprocess(match)

        return match

    def _score_matches(
        self, to_be_matched_instance: str, possible_matches: List
    ) -> np.array:  # type: ignore
        """A method to score a name to_be_matched_instance to a list of possible matches.
        The scoring is done based on all the metrics which are enabled.

        Parameters
        ----------
        to_be_matched_instance : str
            The name which should match one of the possible matches
        possible_matches : List
            list of the names of the possible matches

        Returns
        -------
        np.array
            The score of each of the matches with respect to the different metrics which
            are assessed.
        """
        match_score = np.zeros((len(possible_matches), self._num_distance_metrics))
        idx = 0
        for method_list in self._distance_metrics.values():
            for method in method_list:
                match_score[:, idx] = np.array(
                    [
                        method.sim(str(to_be_matched_instance), str(s))
                        for s in possible_matches
                    ]
                )
                idx = idx + 1

        return match_score

    def _rate_matches(self, match_score: np.array) -> np.array:  # type: ignore
        """Converts the match scores from the score_matches method to a list of indexes of
        the best scoring matches limited to the _number_of_matches.

        Parameters
        ----------
        match_score : np.array
            An array containing the scores of each of the possible alternatives for each
            of the different methods used

        Returns
        -------
        np.array
            The indexes of the best rated matches
        """
        if self._number_of_matches == 1:
            ind = [np.argmax(np.mean(match_score, axis=1))]
        elif self._number_of_matches == len(self._distance_metrics):
            ind = np.zeros(len(self._distance_metrics))
            idx = 0
            for num, method_list in enumerate(self._distance_metrics.values()):
                method_grouped_results = np.reshape(
                    match_score[:, idx : idx + len(method_list)], (-1, len(method_list))
                )
                ind[num] = np.argmax(np.mean(method_grouped_results, axis=1))
                idx = idx + len(method_list)
        elif self._number_of_matches == self._num_distance_metrics:
            ind = np.argmax(match_score, axis=0).reshape(-1)
        else:
            ind = np.argsort(np.mean(match_score, axis=1))[-self._number_of_matches :][
                ::-1
            ]

        return np.array(ind, dtype=int)

    def _get_alternative_names(self, match: pd.Series) -> List:
        """Gets all the possible match names from the match.

        Parameters
        ----------
        match : pd.Series
            The series with the possible matches and original scores

        Returns
        -------
        list
            A list with the alternative names for this match
        """
        alt_names = []

        for num in range(self._number_of_matches):
            alt_names.append(str(match[f"match_name_{num}"]))

        return alt_names

    def _process_words(self, org_name: str, alt_names: List) -> Tuple[str, List]:
        """Removes the words from the word list from the org_name and all the names in
        alt_names .

        Parameters
        ----------
        org_name : str
            The original name for the matching data
        alt_names : list
            A list of names from which the words should be removed

        Returns
        -------
        Tuple[str, list]
            The processed version of the org_name and the alt_names, with the words
            removed
        """
        len_atl_names = len(alt_names)
        for word in self._word_set:
            org_name = " ".join(sub(rf"\b{escape(word)}\b", "", org_name).split())
            for num in range(len_atl_names):
                alt_names[num] = " ".join(
                    sub(rf"\b{escape(word)}\b", "", alt_names[num]).split()
                )

        return org_name, alt_names

    def _adjust_scores(self, match_score: np.array, match: pd.Series) -> pd.Series:  # type: ignore
        """Adjust the scores to be between 0 and 100

        Parameters
        ----------
        match_score : np.array
            An array with the scores for each of the options
        match : pd.Series
            The series with the possible matches and original scores

        Returns
        -------
        pd.Series
            The series with the possible matches and adjusted scores
        """
        for num in range(self._number_of_matches):
            match[f"score_{num}"] = 100 * np.mean(match_score[num, :])

        return match

    def postprocess(self, match: pd.Series) -> pd.Series:
        """Postprocesses the scores to exclude certain specific company words or the
        most common words. In this method only the scores are adjusted, the matches
        still stand.

        Parameters
        ----------
        match : pd.Series
            The series with the possible matches and original scores

        Returns
        -------
        pd.Series
            A new version of the input series with updated scores
        """
        alt_names = self._get_alternative_names(match)
        org_name = str(match["original_name"])

        org_name, alt_names = self._process_words(org_name, alt_names)

        match_score = self._score_matches(org_name, alt_names)
        ind = self._rate_matches(match_score)

        match = self._adjust_scores(match_score[ind, :], match)

        return match

    def _vectorise_data(self, transform: bool = True) -> None:
        """Initialises the TfidfVectorizer, which generates ngrams and weights them
        based on the occurrance. Subsequently the matching data will be used to fit
        the vectoriser and the matching data might also be send to the transform_data
        function depending on the transform boolean.

        Parameters
        ----------
        transform : bool
            A boolean indicating whether or not the data should be transformed after the
            vectoriser is initialised
            default: True
        """
        # self._vec.fit(self._df_matching_data[self._column].values.flatten().astype(str))  # type: ignore
        self._vec.fit(
            self._df_matching_data[self._column].to_numpy()
        )  # .flatten().astype(str))  # type: ignore
        if transform:
            self.transform_data()

    def transform_data(self) -> None:
        """A method which transforms the matching data based on the ngrams transformer.
        After the transformation (the generation of the ngrams), the data is normalised
        by dividing each row by the sum of the row. Subsequently the data is changed to
        a coo sparse matrix format with the column indices in ascending order.
        """
        ngrams = self._vec.transform(self._df_matching_data[self._column].astype(str))
        for i, j in zip(ngrams.indptr[:-1], ngrams.indptr[1:]):
            ngrams.data[i:j] = ngrams.data[i:j] / np.sum(ngrams.data[i:j])
        self._n_grams_matching = ngrams.tocsc()
        if self._low_memory:
            self._n_grams_matching = self._n_grams_matching.tocoo()

    def _search_for_possible_matches(self, to_be_matched: pd.DataFrame) -> np.array:  # type: ignore
        """Generates ngrams from the data which should be matched, calculate the cosine
        simularity between these data and the matching data. Hereafter a top n of the
        matches is selected and returned.

        Parameters
        ----------
        to_be_matched : pd.Series
            A series containing a single instance of the data to be matched

        Returns
        -------
        np.array
            An array of top n values which are most closely matched to the to be matched
            data based on the ngrams
        """
        if self._n_grams_matching is None:
            raise RuntimeError(
                """First the data needs to be transformed to be able to use the sparse """
                + """cosine simularity. To transform the data, run transform_data"""
                + """ or run load_and_process_master_data with transform=True"""
            )

        if self._low_memory:
            results = np.zeros((len(to_be_matched), self._top_n))
            input_data = to_be_matched[self._column_matching]
            for idx, row_name in enumerate(tqdm(input_data, disable=not self._verbose)):
                match_ngrams = self._vec.transform([str(row_name)])
                results[idx, :] = sparse_cosine_top_n(
                    matrix_a=self._n_grams_matching,
                    matrix_b=match_ngrams,
                    top_n=self._top_n,
                    low_memory=self._low_memory,
                    number_of_rows=self._number_of_rows,
                    verbose=self._verbose,
                )
        else:
            match_ngrams = self._vec.transform(
                to_be_matched[self._column_matching].astype(str).tolist()
            ).tocsc()
            results = sparse_cosine_top_n(
                matrix_a=self._n_grams_matching,
                matrix_b=match_ngrams,
                top_n=self._top_n,
                low_memory=self._low_memory,
                number_of_rows=self._number_of_rows,
                verbose=self._verbose,
            )

        return results

    def unicode_to_ascii(self, text: str) -> str:
        """Converts a string to ascii characters trhough transliteration. The
        transliteration map is stored in the transliterations.py file in the
        data folder.

        Parameters
        ----------
        test : str
            The text to be transliterated to ascii characters

        Returns
        -------
        str
            The process text without any non-ascii characters
        """

        normalized_text = normalize("NFD", text)

        return (
            "".join(
                [
                    TRANSLITERATION_MAP.get(char, char)
                    for char in normalized_text
                    if category(char) != "Mn"
                ]
            )
            .encode("ASCII", "ignore")
            .decode()
        )

    def preprocess(
        self, df: pd.DataFrame, column_name: str, original_name: bool = False
    ) -> pd.DataFrame:
        """Preprocess a dataframe before applying a name matching algorithm. The
        preprocessing consists of removing special characters, spaces, converting all
        characters to lower case and removing the words given in the word lists

        Parameters
        ----------
        df : DataFrame
            The dataframe or series on which the preprocessing needs to be performed
        column_name : str
            The name of the column that is used for the preprocessing
        original_name : bool
            If True, returns an additional column 'original_name' in the dataframe
            this column holds the original, non-processed name.
            default=False

        Returns
        -------
        pd.DataFrame
            The preprocessed dataframe or series depending on the input
        """
        if original_name:
            df["original_name"] = df[column_name].copy(deep=True)
        df.loc[:, column_name] = df[column_name].astype(str)
        if self._preprocess_non_word_characters:
            df.loc[:, column_name] = df[column_name].str.replace(
                r"[^\w\-\&\#]", " ", regex=True
            )
            df.loc[:, column_name] = (
                df[column_name].str.replace(r"\s+", " ", regex=True).str.strip()
            )
        if self._preprocess_ascii:
            df.loc[:, column_name] = df[column_name].apply(
                lambda string: self.unicode_to_ascii(str(string))
            )
        if self._preprocess_lowercase:
            df.loc[:, column_name] = df[column_name].str.lower()
        if self._preprocess_legal_suffixes:
            df = self._replace_legal_pre_suffixes_with_abbreviations(df, column_name)
        if self._preprocess_abbreviations:
            df = self._replace_common_strings(df, column_name)
        if self._preprocess_non_word_characters:
            df.loc[:, column_name] = df[column_name].str.replace(
                r"[^\w\-\&\#]", " ", regex=True
            )
            df.loc[:, column_name] = (
                df[column_name].str.replace(r"\s+", " ", regex=True).str.strip()
            )

        return df

    def _preprocess_word_list(self, terms: dict) -> List:
        """Preprocess legal words to remove non-word-characters and trailing leading space

        Parameters
        -------
        terms: dict
            a dictionary of legal words

        Returns
        -------
        list
            A list of preprocessed legal words
        """
        if self._preprocess_non_word_characters:
            return [
                sub(r"[^\w\-\&\#]", "", s).strip()
                for s in reduce(iconcat, terms.values(), [])
            ]
        else:
            return [s.strip() for s in reduce(iconcat, terms.values(), [])]

    def _process_legal_words(self, word_set: set) -> set:
        """Preprocess legal words and add them to the word_set

        Parameters
        -------
        word_set: str
            the current word list which should be extended with additional words

        Returns
        -------
        Set
            The original word_set with the legal words added
        """
        with importlib_resource.as_file(
            importlib_resource.files("name_matching.data").joinpath("legal_names.csv")
        ) as path:
            legal_words = pd.read_csv(path)
        word_set = word_set.union(set(legal_words["abbreviation"].values))

        return word_set

    def _process_common_words(self, word_set: set, cut_off: float) -> set:
        """A method to select the most common words from the matching_data.

        Parameters
        -------
        word_set: str
            the current word list which should be extended with additional words
        cut_off: float
            the cut_off percentage of the occurrence of the most occurring word for
            which words are still included in the no_soring_words set

        Returns
        -------
        Set
            The current word set with the most common words from the matching_data added
        """
        word_counts = (
            self._df_matching_data[self._column]
            .str.split(expand=True)
            .stack()
            .value_counts()
        )
        word_set = word_set.union(
            set(word_counts[word_counts > np.max(word_counts) * cut_off].index)
        )

        return word_set

    def _make_no_scoring_words(
        self, indicator: str, word_set: set, cut_off: float
    ) -> set:
        """A method to make a set of words which are not taken into account when
        scoring matches.

        Parameters
        -------
        indicator: str
            indicator for which types of words should be excluded can be legal for
            legal suffixes or common for the most common words
        word_set: str
            the current word list which should be extended with additional words
        cut_off: float
            the cut_off percentage of the occurrence of the most occurring word for
            which words are still included in the no_soring_words set

        Returns
        -------
        Set
            The set of no scoring words
        """
        if indicator == "legal":
            word_set = self._process_legal_words(word_set)
        if indicator == "common":
            word_set = self._process_common_words(word_set, cut_off)

        return word_set



import copy
import random
from typing import Any, Optional, Protocol

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.base import clone
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.metrics import (
    accuracy_score,
    f1_score,
    precision_recall_curve,
    precision_score,
    recall_score,
)
from sklearn.preprocessing import StandardScaler

from name_matching.match_annotator import MatchAnnotator
from name_matching.name_matcher import NameMatcher


class sklearnModel(Protocol):
    """Protocol for a scikit-learn–style estimator.

    A model must implement the `fit` and `predict` methods.
    """

    def fit(self, X: Any, y: Any) -> Any: ...
    def predict(self, X: Any) -> Any: ...


class scaler(Protocol):
    """Protocol for a scaler.

    A scaler must implement the `fit` and `transform` methods.
    """

    def fit(self, X: Any, y: Any = None) -> Any: ...
    def transform(self, X: Any) -> Any: ...


class NameMatchingOptimiser:
    """Optimises name matching using ML annotation and classification.

    Allows annotating matches, training a classifier on annotated data,
    cross-validation, plotting performance curves, and predicting new matches.
    """

    def __init__(
        self,
        df_matching_data: pd.DataFrame,
        matching_col: str,
        df_to_be_matched: pd.DataFrame,
        to_be_matched_col: str,
        annotated_data: Optional[dict] = None,
        name_matcher: Optional[NameMatcher] = None,
    ) -> None:
        """Initialise with master and target DataFrames and optional annotations.

        Parameters
        ----------
        df_matching_data : pd.DataFrame
            DataFrame containing the master list of names to match against.
        matching_col : str
            Column in `df_matching_data` containing the names.
        df_to_be_matched : pd.DataFrame
            DataFrame containing the names requiring matches.
        to_be_matched_col : str
            Column in `df_to_be_matched` containing those names.
        annotated_data : dict, optional
            Pre-annotated mapping of original names to matched names.
        name_matcher : NameMatcher, optional
            Custom NameMatcher instance. If None, a default is created.

        Attributes
        ----------
        _nm : NameMatcher
            Internal name matcher instance.
        _annotated_data : dict
            Stores annotated results.
        """
        self._df_matching_data = df_matching_data
        self._matching_col = matching_col
        self._df_to_be_matched = df_to_be_matched
        self._to_be_matched_col = to_be_matched_col
        self._annotated_data = annotated_data
        self._features = None
        self._classes = None
        if name_matcher is None:
            self._nm = NameMatcher()
        else:
            self._nm = name_matcher
        self._nm._number_of_matches = self._nm._num_distance_metrics
        self._true_index = None
        self._false_index = None
        self._all_false = None

    def _perform_name_matching(
        self, metrics: Optional[list[str]]
    ) -> pd.DataFrame | pd.Series | tuple[pd.DataFrame, pd.DataFrame]:
        """Perform name matching with specified distance metrics.

        Parameters
        ----------
        metrics : list of str
            List of distance metrics to use (e.g. ['levenshtein', 'jaro']).

        Returns
        -------
        DataFrame or Series or tuple of DataFrames
            If multiple distances are returned separately, returns a tuple;
            otherwise returns a DataFrame or Series of match scores.

        Raises
        ------
        ValueError
            If `metrics` is None or empty.
        """
        if metrics is None:
            raise ValueError("No metrics supplied!")
        self._nm.load_and_process_master_data(
            self._matching_col, self._df_matching_data, False
        )
        self._nm.set_distance_metrics(metrics)
        matches = self._nm.match_names(self._df_to_be_matched, self._to_be_matched_col)
        return matches

    def _annotate(self, matches: pd.DataFrame, lower_bound: float) -> None:
        """Launch manual annotation on a slice of matches above a threshold.

        Parameters
        ----------
        matches : pd.DataFrame
            DataFrame containing candidate matches with score columns.
        lower_bound : float
            Minimum score (%) for considering a pair for annotation.
        """
        if self._annotated_data is None:
            self._annotated_data = {}
        filtered = self._preselect_matches(matches, lower_bound)
        rc = MatchAnnotator(filtered, annotated_results=self._annotated_data)
        rc.start()
        self._annotated_data.update(rc.annotated_results)

    def _annotate_matches(self, data: pd.DataFrame) -> None:
        """Automatically annotate each original name with its highest-score match.

        Parameters
        ----------
        data : pd.DataFrame
            DataFrame of matches with one column for each score and
            columns 'original_name' and 'match_name_X'.
        """
        names = data["original_name"]
        score_cols = data.columns.str.contains("score")
        max_col = data.loc[:, score_cols].idxmax(axis=1)
        max_col = max_col.str.split("_").apply(lambda parts: "match_name_" + parts[1])
        for orig, idx, best in zip(names.values, names.index, max_col.values):
            self._annotated_data[orig] = data.loc[idx, best]  # type: ignore

    @staticmethod
    def split_list_random_uneven(lst: list[Any], x: int) -> list[list[Any]]:
        """Shuffle `lst` and split into `x` roughly equal parts.

        Elements are randomly distributed; remainders are distributed one by one
        to the first few sublists.

        Parameters
        ----------
        lst : list
            Input list to split.
        x : int
            Number of sublists to return.

        Returns
        -------
        list of lists
            A list containing `x` sublists of `lst`.
        """
        lst_copy = copy.deepcopy(lst)
        random.shuffle(lst_copy)
        k, m = divmod(len(lst_copy), x)
        return [
            lst_copy[i * k + min(i, m) : (i + 1) * k + min(i + 1, m)] for i in range(x)
        ]

    def cross_validate_model(
        self, model_name: Optional[str] = None, threshold: float = 0.5, n_folds: int = 5
    ) -> pd.DataFrame:
        """Perform `n_folds` cross-validation on the fitted model.

        Parameters
        ----------
        model_name : str, optional
            Name to assign to the model in the results table.
        threshold : float, default=0.5
            Probability threshold for binary decisions.
        n_folds : int, default=5
            Number of folds for random splitting.

        Returns
        -------
        pd.DataFrame
            MultiIndex DataFrame containing mean and std for accuracy,
            precision, recall, and F1 across folds.

        Raises
        ------
        ValueError
            If the model has not been fitted (no `self._true_index`).
        """
        if self._true_index is None:
            raise ValueError("Please fit the model before running cross-validation.")

        accuracy, recall, precision, f1 = [], [], [], []
        lists = self.split_list_random_uneven(list(self._true_index.keys()), n_folds)

        for train_idx in lists:
            X_train, y_train, X_test, y_test = self._generate_balanced_split(
                0.8, train_idx
            )
            model = clone(self.model)
            model.fit(X_train, y_train)

            if hasattr(model, "predict_proba"):
                probs = model.predict_proba(X_test)[:, 1]
                y_pred = (probs >= threshold).astype(int)
            else:
                y_pred = model.predict(X_test)

            accuracy.append(accuracy_score(y_test, y_pred))
            recall.append(recall_score(y_test, y_pred))
            precision.append(precision_score(y_test, y_pred))
            f1.append(f1_score(y_test, y_pred))

        if model_name is None:
            model_name = type(self.model).__name__
        index = pd.MultiIndex.from_tuples(
            [
                ("Accuracy", "μ"),
                ("Accuracy", "σ"),
                ("F1", "μ"),
                ("F1", "σ"),
                ("Precision", "μ"),
                ("Precision", "σ"),
                ("Recall", "μ"),
                ("Recall", "σ"),
            ]
        )
        df = pd.DataFrame(index=index, columns=[model_name])
        df.loc[("Accuracy", "μ"), model_name] = np.mean(accuracy)
        df.loc[("Accuracy", "σ"), model_name] = np.std(accuracy)
        df.loc[("F1", "μ"), model_name] = np.mean(f1)
        df.loc[("F1", "σ"), model_name] = np.std(f1)
        df.loc[("Precision", "μ"), model_name] = np.mean(precision)
        df.loc[("Precision", "σ"), model_name] = np.std(precision)
        df.loc[("Recall", "μ"), model_name] = np.mean(recall)
        df.loc[("Recall", "σ"), model_name] = np.std(recall)

        return df

    def _preselect_matches(
        self, matches: pd.DataFrame, lower_bound: float
    ) -> pd.DataFrame:
        """Filter matches to those with max score between `lower_bound` and 100.

        Parameters
        ----------
        matches : pd.DataFrame
            All candidate matches with one or more ‘score’ columns.
        lower_bound : float
            Minimum acceptable score (exclusive).

        Returns
        -------
        pd.DataFrame
            Subset of `matches` with `lower_bound < max(score_cols) < 100`.
        """
        score_mask = matches.columns.str.contains("score")
        matches["max_scr"] = matches.loc[:, score_mask].max(axis=1)
        return matches[(matches["max_scr"] > lower_bound) & (matches["max_scr"] < 100)]

    def annotate(
        self,
        lower_bound: float = 70.0,
        data_percentage: float = 0.2,
        max_matches: int = 10,
    ) -> None:
        """Sample a subset of the target names and manually annotate matches.

        Parameters
        ----------
        lower_bound : float, default=70.0
            Minimum match score (%) for prompting annotation.
        data_percentage : float, default=0.2
            Fraction of `df_to_be_matched` to sample for annotation.
        max_matches : int, default=10
            Maximum number of candidate matches per name to present.
        """
        self._nm._return_algorithms_score = False
        max_matches = min(max_matches, self._nm._top_n)
        self._nm._number_of_matches = max_matches
        self._nm.load_and_process_master_data(
            self._matching_col, self._df_matching_data, False
        )
        reduced = self._df_to_be_matched.sample(frac=data_percentage, replace=False)
        candidates = self._nm.match_names(reduced, self._to_be_matched_col)
        self._annotate(candidates, lower_bound)  # type: ignore

    def export_annotation(
        self, file_name: Optional[str] = None
    ) -> Optional[pd.DataFrame]:
        """Export the current annotations to CSV or return as DataFrame.

        Parameters
        ----------
        file_name : str, optional
            Path to save the CSV. If None, returns a DataFrame instead.

        Returns
        -------
        pd.DataFrame or None
            If `file_name` is None, returns a DataFrame of annotations.

        Raises
        ------
        ValueError
            If no annotations exist.
        """
        if not isinstance(self._annotated_data, dict):
            raise ValueError("No annotated data available.")

        df = pd.DataFrame(
            {
                "original_name": list(self._annotated_data.keys()),
                "match_name": list(self._annotated_data.values()),
            }
        )
        if file_name:
            df.to_csv(file_name, index=False)
            return None
        return df

    def import_annotation(self, annotations: pd.DataFrame | str) -> None:
        """Load annotations from a DataFrame or a CSV file.

        Parameters
        ----------
        annotations : pd.DataFrame or str
            If DataFrame, must have columns `original_name` and `match_name`.
            If str, interpreted as CSV file path.
        """
        if isinstance(annotations, str):
            annotations = pd.read_csv(annotations)
        ann_dict = annotations.set_index("original_name")["match_name"].to_dict()
        if self._annotated_data is None:
            self._annotated_data = ann_dict
        else:
            self._annotated_data.update(ann_dict)

    def plot_curves(
        self, absolute: bool = True, return_data: bool = False
    ) -> None | dict[str, Any]:
        """Plot Precision-Recall curve and threshold-based TP/FP/ TN/FN curves.

        Parameters
        ----------
        absolute : bool, default=True
            If True, plot absolute counts; if False, plot normalized rates.
        return_data : bool, default=False
            If True, return the underlying curve data.

        Returns
        -------
        dict or None
            If `return_data` is True, returns a dict with arrays for
            ‘true_positives’, ‘false_positives’, ‘true_negatives’,
            ‘false_negatives’, ‘threshold’, ‘Precision’, and ‘Recall’.
        """
        y_scores = self.model.predict_proba(self._X_test)[:, 1]
        precision, recall, _ = precision_recall_curve(self._y_test, y_scores)

        plt.figure(figsize=(8, 6))
        plt.plot(recall, precision, label="Precision-Recall curve")
        plt.xlabel("Recall")
        plt.ylabel("Precision")
        plt.title("Precision-Recall Curve")
        plt.legend()
        plt.show()

        thresholds = np.linspace(0, 1, 1000)
        tp, fp, tn, fn = [], [], [], []
        for t in thresholds:
            pos = y_scores >= t
            neg = ~pos
            tp.append(np.sum(pos & (self._y_test == 1)))
            fp.append(np.sum(pos & (self._y_test == 0)))
            tn.append(np.sum(neg & (self._y_test == 0)))
            fn.append(np.sum(neg & (self._y_test == 1)))
            if not absolute:
                total_pos = np.sum(self._y_test == 1)
                total_neg = np.sum(self._y_test == 0)
                tp[-1] /= total_pos
                fp[-1] /= total_neg
                tn[-1] /= total_neg
                fn[-1] /= total_pos

        plt.figure(figsize=(8, 6))
        plt.plot(thresholds, tp, label="True positives")
        plt.plot(thresholds, fp, label="False positives")
        plt.xlabel("Threshold")
        plt.ylabel("Count" if absolute else "Rate")
        plt.ylim(0, (np.sum(self._y_test == 1) if absolute else 1))
        plt.legend()
        plt.show()

        if return_data:
            return {
                "true_positives": tp,
                "false_positives": fp,
                "true_negatives": tn,
                "false_negatives": fn,
                "threshold": thresholds,
                "Precision": precision,
                "Recall": recall,
            }

    def _preprocess_fit_annotations(self, dataScaler: type[scaler]) -> None:
        """Prepare training and test indices from annotated matches.

        Runs name matching on all data, then splits each annotated example
        into one positive and one or more negative feature vectors.

        Parameters
        ----------
        dataScaler : class
            Scaler class (e.g. StandardScaler) to fit on all feature vectors.

        Raises
        ------
        ValueError
            If no annotated data is available.
        """
        self._true_index = {}
        self._false_index = {}
        self._all_false = {}

        if not isinstance(self._annotated_data, dict):
            raise ValueError("Annotated data required before fitting.")

        # Generate all candidate scores and names
        self._nm._return_algorithms_score = True
        self._nm.load_and_process_master_data(
            self._matching_col, self._df_matching_data, False
        )
        matches, possible = self._nm.match_names(
            self._df_to_be_matched, self._to_be_matched_col
        )
        matches = matches.reset_index(drop=True)  # type: ignore

        # Fit scaler on full set
        self.scaler = dataScaler().fit(matches[0])  # type: ignore

        idx_true = idx_false = 0
        for idx, name in enumerate(self._df_to_be_matched[self._to_be_matched_col]):
            if name in self._annotated_data:
                true_match = self._annotated_data[name]
                pos_mask = possible[idx] == true_match  # type: ignore
                if np.any(pos_mask):
                    self._true_index[idx_true] = matches[idx][pos_mask][0]  # type: ignore
                    all_other = np.delete(matches[idx], pos_mask, axis=0)  # type: ignore
                    self._all_false[idx_true] = all_other
                    self._false_index[idx_true] = all_other[np.argmax(all_other.mean(axis=1))]  # type: ignore
                    idx_true += 1
                elif true_match == -1:
                    best = matches[idx][np.argmax(matches[idx].mean(axis=1))]  # type: ignore
                    self._false_index[idx_false] = best
                    idx_false += 1

    def _generate_balanced_split(
        self, train_split: float, train_idx: Optional[list[int]] = None
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Build balanced train/test sets from true and false example indices.

        For each annotated index, pairs a true vector with one false vector.

        Parameters
        ----------
        train_split : float
            Fraction of `self._true_index` keys to include in training.
        train_idx : list of int, optional
            Specific keys to use for training; if None, chosen randomly.

        Returns
        -------
        X_train : ndarray
        y_train : ndarray
        X_test : ndarray
        y_test : ndarray

        Raises
        ------
        ValueError
            If `self._true_index`, `self._false_index`, or `self._all_false` is None.
        """
        if (
            self._true_index is None
            or self._false_index is None
            or self._all_false is None
        ):
            raise ValueError("Fit preprocessing must be run first.")

        keys = list(self._true_index.keys())
        if train_idx is None:
            n_train = int(train_split * len(keys))
            train_idx = list(np.random.choice(keys, n_train, replace=False))
        test_idx = [k for k in keys if k not in train_idx]

        # Build training arrays
        X_train = np.vstack(
            [self._true_index[i] for i in train_idx]
            + [self._false_index[i] for i in train_idx]
        )
        y_train = np.hstack([np.ones(len(train_idx)), np.zeros(len(train_idx))])

        # Build test arrays
        X_test = np.vstack(
            [self._true_index[i] for i in test_idx]
            + [self._all_false[i] for i in test_idx]
        )
        y_test = np.hstack(
            [
                np.ones(len(test_idx)),
                np.zeros(sum(len(self._all_false[i]) for i in test_idx)),
            ]
        )

        X_train = self.scaler.transform(X_train)
        X_test = self.scaler.transform(X_test)

        return X_train, y_train, X_test, y_test

    def fit(
        self,
        model: sklearnModel = GradientBoostingClassifier,
        dataScaler: scaler = StandardScaler,
        model_args: Optional[dict] = None,
        print_results: bool = True,
        train_split: float = 0.7,
        preprocess: bool = False,
    ) -> None:
        """Fit a classifier on annotated name-matching examples.

        This will preprocess annotations (if needed), split into train/test,
        scale features, and fit the chosen model.

        Parameters
        ----------
        model : estimator class, default=GradientBoostingClassifier
            Scikit-learn compatible classifier to fit.
        dataScaler : scaler class, default=StandardScaler
            Scaler to normalise feature vectors.
        model_args : dict, optional
            Keyword arguments to pass to the model constructor.
        print_results : bool, default=True
            If True, print accuracy, recall, and precision on the test set.
        train_split : float, default=0.7
            Fraction of data to allocate to training vs. testing.
        preprocess : bool, default=False
            If True or if no prior preprocessing, rerun annotation preprocessing.
        """
        self._model = model
        if preprocess or self._true_index is None:
            self._preprocess_fit_annotations(dataScaler)

        self._X_train, self._y_train, self._X_test, self._y_test = (
            self._generate_balanced_split(train_split)
        )

        self._model_args = model_args or {}
        self.model = self._model(**self._model_args)  # type: ignore
        self.model.fit(self._X_train, self._y_train)

        if print_results:
            y_pred = self.model.predict(self._X_test)
            print(f"accuracy - {accuracy_score(self._y_test, y_pred)}")
            print(f"recall   - {recall_score(self._y_test, y_pred)}")
            print(f"precision- {precision_score(self._y_test, y_pred)}")

    def predict(self, threshold: float = 0.5) -> dict[str, str]:
        """Match new names by applying the trained classifier to all candidates.

        Parameters
        ----------
        threshold : float, default=0.5
            Probability cutoff above which a candidate is accepted.

        Returns
        -------
        dict
            Mapping from each original name in `df_to_be_matched` to the
            best match name (or empty string if below threshold).
        """
        self._nm._return_algorithms_score = True
        self._nm.load_and_process_master_data(
            self._matching_col, self._df_matching_data, False
        )
        scores, names = self._nm.match_names(
            self._df_to_be_matched, self._to_be_matched_col
        )
        feats = np.stack(scores.to_numpy()).reshape(-1, self._nm._num_distance_metrics)
        feats = self.scaler.transform(feats)
        probs = self.model.predict_proba(feats)[:, 1]
        probs = probs.reshape(-1, self._nm._top_n)

        results: dict[str, str] = {}
        for i, orig in enumerate(self._df_to_be_matched[self._to_be_matched_col]):
            best = np.argmax(probs[i])
            results[orig] = names[i, best] if probs[i, best] > threshold else ""
        return results



# Copyright 2018-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.tokenizer._tokenize.

_Tokenizer base class
"""

from collections import Counter, defaultdict
from math import exp, log1p, log2
from typing import Any, Callable
from typing import Counter as TCounter
from typing import List, Optional, Set, Union, cast

__all__ = ["_Tokenizer"]


class _Tokenizer:
    """Abstract _Tokenizer class.

    .. versionadded:: 0.4.0
    """

    def __init__(
        self,
        scaler: Optional[Union[str, Callable[[float], float]]] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Initialize Tokenizer.

        Parameters
        ----------
        scaler : None, str, or function
            A scaling function for the Counter:

                - None : no scaling
                - 'set' : All non-zero values are set to 1.
                - 'length' : Each token has weight equal to its length.
                - 'length-log' : Each token has weight equal to the log of its
                   length + 1.
                - 'length-exp' : Each token has weight equal to e raised to its
                   length.
                - 'entropy' : Weights are scaled to the (log_2) information
                  entropy of each key's frequency.
                - a callable function : The function is applied to each value
                  in the Counter. Some useful functions include math.exp,
                  math.log1p, math.sqrt, and indexes into interesting integer
                  sequences such as the Fibonacci sequence.


        .. versionadded:: 0.4.0

        """
        super(_Tokenizer, self).__init__()

        self._scaler = scaler
        self._tokens = defaultdict(int)  # type: DefaultDict[str, float]
        self._string = ""
        self._ordered_tokens = []  # type: List[str]
        self._ordered_weights = []  # type: List[float]

    def tokenize(self, string: str) -> "_Tokenizer":
        """Tokenize the term and store it.

        The tokenized term is stored as an ordered list and as a defaultdict
        object.

        Parameters
        ----------
        string : str
            The string to tokenize


        .. versionadded:: 0.4.0
        .. versionchanged:: 0.4.1
            Added 'length', 'entropy', and related scalers
        .. versionchanged:: 0.6.0
            Moved scaling & counterizing to separate function

        """
        self._string = string
        self._ordered_tokens = [self._string]
        self._ordered_weights = [1]

        self._scale_and_counterize()
        return self

    def _scale_and_counterize(self) -> None:
        """Scale the tokens and store them in a defaultdict.

        .. versionadded:: 0.6.0

        """
        if self._scaler in {"SSK", "length", "length-log", "length-exp"}:
            self._tokens = defaultdict(float)
            if cast(str, self._scaler)[:6] == "length":
                self._ordered_weights = [len(_) for _ in self._ordered_tokens]
                if self._scaler == "length-log":
                    self._ordered_weights = [log1p(_) for _ in self._ordered_weights]
                elif self._scaler == "length-exp":
                    self._ordered_weights = [exp(_) for _ in self._ordered_weights]
            for token, weight in zip(self._ordered_tokens, self._ordered_weights):
                self._tokens[token] += weight
        elif self._scaler == "entropy":
            counts = Counter(self._ordered_tokens)
            n = len(self._ordered_tokens)
            self._tokens = defaultdict(float)
            self._tokens.update(
                {key: -(val / n) * log2(val / n) for key, val in counts.items()}
            )
            self._ordered_weights = [
                self._tokens[tok] / counts[tok] for tok in self._ordered_tokens
            ]
        else:
            self._tokens = defaultdict(int)
            self._tokens.update(Counter(self._ordered_tokens))

    def count(self) -> int:
        """Return token count.

        Returns
        -------
        int
            The total count of tokens

        Examples
        --------
        >>> tok = _Tokenizer().tokenize('term')
        >>> tok.count()
        1


        .. versionadded:: 0.4.0

        """
        return sum(self.get_counter().values())

    def count_unique(self) -> int:
        """Return the number of unique elements.

        Returns
        -------
        int
            The number of unique tokens

        Examples
        --------
        >>> tok = _Tokenizer().tokenize('term')
        >>> tok.count_unique()
        1


        .. versionadded:: 0.4.0

        """
        return len(self._tokens.values())

    def get_counter(self) -> TCounter[str]:
        """Return the tokens as a Counter object.

        Returns
        -------
        Counter
            The Counter of tokens

        Examples
        --------
        >>> tok = _Tokenizer().tokenize('term')
        >>> tok.get_counter()
        Counter({'term': 1})


        .. versionadded:: 0.4.0

        """
        if self._scaler == "set":
            return Counter({key: 1 for key in self._tokens.keys()})
        elif callable(self._scaler):
            return Counter(
                {key: self._scaler(val) for key, val in self._tokens.items()}
            )
        else:
            return Counter(self._tokens)

    def get_set(self) -> Set[str]:
        """Return the unique tokens as a set.

        Returns
        -------
        Counter
            The set of tokens

        Examples
        --------
        >>> tok = _Tokenizer().tokenize('term')
        >>> tok.get_set()
        {'term'}


        .. versionadded:: 0.4.0

        """
        return set(self._tokens.keys())

    def get_list(self) -> List[str]:
        """Return the tokens as an ordered list.

        Returns
        -------
        Counter
            The list of q-grams in the order they were added.

        Examples
        --------
        >>> tok = _Tokenizer().tokenize('term')
        >>> tok.get_list()
        ['term']


        .. versionadded:: 0.4.0

        """
        return self._ordered_tokens

    def __repr__(self) -> str:
        """Return representation of tokens object.

        .. versionadded:: 0.4.0

        """
        return self.__class__.__name__ + "({}".format(str(self._tokens)[27:])

    def __and__(self, other: "_Tokenizer") -> TCounter[str]:
        """Return intersection with other tokens.

        .. versionadded:: 0.4.0

        """
        return self.get_counter() & other.get_counter()

    def __add__(self, other: "_Tokenizer") -> TCounter[str]:
        """Return union with other tokens.

        .. versionadded:: 0.4.0

        """
        return self.get_counter() + other.get_counter()

    def __sub__(self, other: "_Tokenizer") -> TCounter[str]:
        """Return difference from other tokens.

        .. versionadded:: 0.4.0

        """
        return self.get_counter() - other.get_counter()


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import ipywidgets as ipyw
import numpy as np
import pandas as pd
from IPython.display import clear_output, display


class MatchAnnotator:
    """A class to annotate matches. The annotated matches can be used to automate the
    decision on whether something is a correct match or not."""

    def __init__(
        self,
        results: pd.DataFrame,
        name_col: str = "original_name",
        annotated_results: dict | None = None,
    ) -> None:
        """Initialize the MatchAnnotator with results and settings.

        Parameters
        ----------
        results : type
            Description of results.
        name_col : type
            Description of name_col.
        annotated_results : type
            Description of annotated_results.
        """
        self._results_data = results
        self._name_col = name_col
        if annotated_results is None:
            self._annotated_results = {}
        else:
            self._annotated_results = annotated_results

    @property
    def annotated_results(self) -> dict:
        """Return the dictionary of annotated results.


        Returns
        -------
        dict
            Description of return value.
        """
        return self._annotated_results

    def _initiate_buttons(self) -> list:
        """Initialize and return annotation control buttons.


        Returns
        -------
        list
            Description of return value.
        """
        button_nm = ipyw.Button(
            description="no match", layout=ipyw.Layout(width="100%")
        )
        button_nm.on_click(self._no_match)
        button_dn = ipyw.Button(
            description="don't know", layout=ipyw.Layout(width="100%")
        )
        button_dn.on_click(self._skip)
        button_stop = ipyw.Button(
            description="stop",
            style={"button_color": "white"},
            layout=ipyw.Layout(width="100%"),
        )
        button_stop.on_click(self._stop)

        return [button_nm, button_dn, button_stop]

    def _no_match(self, button: ipyw.Button) -> None:
        """Handle 'no match' button click and skip to next item.

        Parameters
        ----------
        button : ipyw.Button
            Description of button.
        """
        self._annotated_results[
            self._results_data.loc[self._possible_nodes[self._index], "original_name"]
        ] = -1
        self._skip(None)

    def _possible_names(self) -> np.ndarray:
        """Retrieve possible match names for the current item.

        Returns
        -------
        np.ndarray : type
            Description of return value.
        """
        row = self._results_data.loc[self._possible_nodes[self._index]]
        return row[row.index.str.contains("match_name")].unique()

    def _skip(self, button: ipyw.Button | None) -> None:
        """Skip to the next item in the annotation list.

        Parameters
        ----------
        button : type
            Description of button.
        """
        self._index = self._index + 1
        if self._index >= len(self._possible_nodes):
            self._stop(None)
        else:
            buttons = self._initiate_buttons()
            for name in self._possible_names():
                self._add_button(buttons, name)
            self._show_data(buttons)

    def _save_result(self, button: ipyw.Button) -> None:
        """Save the selected match result and skip to next item.

        Parameters
        ----------
        button : type
            Description of button.
        """
        self._annotated_results[
            self._results_data.loc[self._possible_nodes[self._index], "original_name"]
        ] = button.description
        self._skip(None)

    def _add_button(self, buttons: list, name: str) -> None:
        """Add a match name button to the list of buttons.

        Parameters
        ----------
        buttons : type
            Description of buttons.
        name : type
            Description of name.
        """
        button = ipyw.Button(description=name, layout=ipyw.Layout(width="100%"))
        button.on_click(self._save_result)
        buttons.insert(1, button)

    def _show_data(self, buttons: list) -> None:
        """Display the current item and annotation buttons.

        Parameters
        ----------
        buttons : type
            Description of buttons.
        """
        clear_output()
        label = ipyw.HTML(
            value=(
                "<b style='font-size:20px'>"
                f"{self._results_data.loc[self._possible_nodes[self._index],'original_name']}"
                "</b>"
            ),
            layout=ipyw.Layout(
                padding="10px",
                width="100%",
                height="50px",
            ),
        )
        buttons.insert(0, label)
        display(ipyw.VBox(children=buttons, layout=ipyw.Layout(width="500px")))

    def _stop(self, button: ipyw.Button) -> None:
        """Stop the annotation process and clear output.

        Parameters
        ----------
        button : type
            Description of button.
        """
        clear_output()

    def _define_possible_nodes(self, selection) -> list:
        """Define the list of items to annotate based on selection.

        Parameters
        ----------
        selection : type
            Description of selection.

        Returns
        -------
        list : type
            Description of return value.
        """

        if selection == "random":
            return (
                self._results_data.index.to_series()
                .sample(n=len(self._results_data))
                .to_list()
            )

    def start(self, selection: str = "random", fresh_start: bool = False) -> None:
        """Start the annotation process with optional fresh start.

        Parameters
        ----------
        selection : type
            Description of selection.
        fresh_start : type
            Description of fresh_start.
        """
        self._index = -1
        self._possible_nodes = self._define_possible_nodes(selection)
        if fresh_start:
            self._annotated_results = {}
        self._skip(None)



# Copyright 2014-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.phonetic._double_metaphone.

Double Metaphone
"""

from typing import Set, Tuple

from ._phonetic import _Phonetic

__all__ = ["DoubleMetaphone"]


class DoubleMetaphone(_Phonetic):
    """Double Metaphone.

    Based on Lawrence Philips' (Visual) C++ code from 1999
    :cite:`Philips:2000`.

    .. versionadded:: 0.3.6
    """

    def __init__(self, max_length: int = -1) -> None:
        """Initialize DoubleMetaphone instance.

        Parameters
        ----------
        max_length : int
            Maximum length of the returned Dolby code -- this also activates
            the fixed-length code mode if it is greater than 0


        .. versionadded:: 0.4.0

        """
        self._max_length = max_length

        # Require a max_length of at least 4
        if self._max_length != -1:
            self._max_length = max(4, max_length)

    def encode_alpha(self, word: str) -> str:
        """Return the alphabetic Double Metaphone code for a word.

        Parameters
        ----------
        word : str
            The word to transform

        Returns
        -------
        str
            The alphabetic Double Metaphone value(s)

        Examples
        --------
        >>> pe = DoubleMetaphone()
        >>> pe.encode_alpha('Christopher')
        'KRSTFR,'
        >>> pe.encode_alpha('Niall')
        'NL,'
        >>> pe.encode_alpha('Smith')
        'SMÞ,XMT'
        >>> pe.encode_alpha('Schmidt')
        'XMT,SMT'


        .. versionadded:: 0.4.0
        .. versionchanged:: 0.6.0
            Made return a str only (comma-separated)

        """
        return self.encode(word).replace("0", "Þ")

    def encode(self, word: str) -> str:
        """Return the Double Metaphone code for a word.

        Parameters
        ----------
        word : str
            The word to transform

        Returns
        -------
        str
            The Double Metaphone value(s)

        Examples
        --------
        >>> pe = DoubleMetaphone()
        >>> pe.encode('Christopher')
        'KRSTFR,'
        >>> pe.encode('Niall')
        'NL,'
        >>> pe.encode('Smith')
        'SM0,XMT'
        >>> pe.encode('Schmidt')
        'XMT,SMT'


        .. versionadded:: 0.1.0
        .. versionchanged:: 0.3.6
            Encapsulated in class
        .. versionchanged:: 0.6.0
            Made return a str only (comma-separated)

        """
        primary = ""
        secondary = ""

        def _slavo_germanic() -> bool:
            """Return True if the word appears to be Slavic or Germanic.

            Returns
            -------
            bool
                True if the word appears to be Slavic or Germanic

            .. versionadded:: 0.1.0

            """
            if "W" in word or "K" in word or "CZ" in word:
                return True
            return False

        def _metaph_add(pri: str, sec: str = "") -> Tuple[str, str]:
            """Return a new metaphone tuple with the supplied elements.

            Parameters
            ----------
            pri : str
                The primary element
            sec : str
                The secondary element

            Returns
            -------
            tuple
                A new metaphone tuple with the supplied elements

            .. versionadded:: 0.1.0

            """
            newpri = primary
            newsec = secondary
            if pri:
                newpri += pri
            if sec:
                if sec != " ":
                    newsec += sec
            else:
                newsec += pri
            return newpri, newsec

        def _is_vowel(pos: int) -> bool:
            """Return True if the character at word[pos] is a vowel.

            Parameters
            ----------
            pos : int
                Position in the word

            Returns
            -------
            bool
                True if the character is a vowel

            .. versionadded:: 0.1.0

            """
            if pos >= 0 and word[pos] in {"A", "E", "I", "O", "U", "Y"}:
                return True
            return False

        def _get_at(pos: int) -> str:
            """Return the character at word[pos].

            Parameters
            ----------
            pos : int
                Position in the word

            Returns
            -------
            str
                Character at word[pos]

            .. versionadded:: 0.1.0

            """
            return word[pos]

        def _string_at(pos: int, slen: int, substrings: Set[str]) -> bool:
            """Return True if word[pos:pos+slen] is in substrings.

            Parameters
            ----------
            pos : int
                Position in the word
            slen : int
                Substring length
            substrings : set
                Substrings to search

            Returns
            -------
            bool
                True if word[pos:pos+slen] is in substrings

            .. versionadded:: 0.1.0

            """
            if pos < 0:
                return False
            return word[pos : pos + slen] in substrings

        current = 0
        length = len(word)
        if length < 1:
            return ","
        last = length - 1

        word = word.upper()

        # Pad the original string so that we can index beyond the edge of the
        # world
        word += "     "

        # Skip these when at start of word
        if word[0:2] in {"GN", "KN", "PN", "WR", "PS"}:
            current += 1

        # Initial 'X' is pronounced 'Z' e.g. 'Xavier'
        if _get_at(0) == "X":
            primary, secondary = _metaph_add("S")  # 'Z' maps to 'S'
            current += 1

        # Main loop
        while True:
            if current >= length:
                break

            if _get_at(current) in {"A", "E", "I", "O", "U", "Y"}:
                if current == 0:
                    # All init vowels now map to 'A'
                    primary, secondary = _metaph_add("A")
                current += 1
                continue

            elif _get_at(current) == "B":
                # "-mb", e.g", "dumb", already skipped over...
                primary, secondary = _metaph_add("P")
                if _get_at(current + 1) == "B":
                    current += 2
                else:
                    current += 1
                continue

            elif _get_at(current) == "Ç":
                primary, secondary = _metaph_add("S")
                current += 1
                continue

            elif _get_at(current) == "C":
                # Various Germanic
                if (
                    current > 1
                    and not _is_vowel(current - 2)
                    and _string_at((current - 1), 3, {"ACH"})
                    and (
                        (_get_at(current + 2) != "I")
                        and (
                            (_get_at(current + 2) != "E")
                            or _string_at((current - 2), 6, {"BACHER", "MACHER"})
                        )
                    )
                ):
                    primary, secondary = _metaph_add("K")
                    current += 2
                    continue

                # Special case 'caesar'
                elif current == 0 and _string_at(current, 6, {"CAESAR"}):
                    primary, secondary = _metaph_add("S")
                    current += 2
                    continue

                # Italian 'chianti'
                elif _string_at(current, 4, {"CHIA"}):
                    primary, secondary = _metaph_add("K")
                    current += 2
                    continue

                elif _string_at(current, 2, {"CH"}):
                    # Find 'Michael'
                    if current > 0 and _string_at(current, 4, {"CHAE"}):
                        primary, secondary = _metaph_add("K", "X")
                        current += 2
                        continue

                    # Greek roots e.g. 'chemistry', 'chorus'
                    elif (
                        current == 0
                        and (
                            _string_at((current + 1), 5, {"HARAC", "HARIS"})
                            or _string_at(
                                (current + 1), 3, {"HOR", "HYM", "HIA", "HEM"}
                            )
                        )
                        and not _string_at(0, 5, {"CHORE"})
                    ):
                        primary, secondary = _metaph_add("K")
                        current += 2
                        continue

                    # Germanic, Greek, or otherwise 'ch' for 'kh' sound
                    elif (
                        (
                            _string_at(0, 4, {"VAN ", "VON "})
                            or _string_at(0, 3, {"SCH"})
                        )
                        # 'architect but not 'arch', 'orchestra', 'orchid'
                        or _string_at((current - 2), 6, {"ORCHES", "ARCHIT", "ORCHID"})
                        or _string_at((current + 2), 1, {"T", "S"})
                        or (
                            (
                                _string_at((current - 1), 1, {"A", "O", "U", "E"})
                                or (current == 0)
                            )
                            # e.g., 'wachtler', 'wechsler', but not 'tichner'
                            and _string_at(
                                (current + 2),
                                1,
                                {
                                    "L",
                                    "R",
                                    "N",
                                    "M",
                                    "B",
                                    "H",
                                    "F",
                                    "V",
                                    "W",
                                    " ",
                                },
                            )
                        )
                    ):
                        primary, secondary = _metaph_add("K")

                    else:
                        if current > 0:
                            if _string_at(0, 2, {"MC"}):
                                # e.g., "McHugh"
                                primary, secondary = _metaph_add("K")
                            else:
                                primary, secondary = _metaph_add("X", "K")
                        else:
                            primary, secondary = _metaph_add("X")

                    current += 2
                    continue

                # e.g, 'czerny'
                elif _string_at(current, 2, {"CZ"}) and not _string_at(
                    (current - 2), 4, {"WICZ"}
                ):
                    primary, secondary = _metaph_add("S", "X")
                    current += 2
                    continue

                # e.g., 'focaccia'
                elif _string_at((current + 1), 3, {"CIA"}):
                    primary, secondary = _metaph_add("X")
                    current += 3

                # double 'C', but not if e.g. 'McClellan'
                elif _string_at(current, 2, {"CC"}) and not (
                    (current == 1) and (_get_at(0) == "M")
                ):
                    # 'bellocchio' but not 'bacchus'
                    if _string_at((current + 2), 1, {"I", "E", "H"}) and not _string_at(
                        (current + 2), 2, {"HU"}
                    ):
                        # 'accident', 'accede' 'succeed'
                        if (
                            (current == 1) and _get_at(current - 1) == "A"
                        ) or _string_at((current - 1), 5, {"UCCEE", "UCCES"}):
                            primary, secondary = _metaph_add("KS")
                        # 'bacci', 'bertucci', other italian
                        else:
                            primary, secondary = _metaph_add("X")
                        current += 3
                        continue
                    else:  # Pierce's rule
                        primary, secondary = _metaph_add("K")
                        current += 2
                        continue

                elif _string_at(current, 2, {"CK", "CG", "CQ"}):
                    primary, secondary = _metaph_add("K")
                    current += 2
                    continue

                elif _string_at(current, 2, {"CI", "CE", "CY"}):
                    # Italian vs. English
                    if _string_at(current, 3, {"CIO", "CIE", "CIA"}):
                        primary, secondary = _metaph_add("S", "X")
                    else:
                        primary, secondary = _metaph_add("S")
                    current += 2
                    continue

                # else
                else:
                    primary, secondary = _metaph_add("K")

                    # name sent in 'mac caffrey', 'mac gregor
                    if _string_at((current + 1), 2, {" C", " Q", " G"}):
                        current += 3
                    elif _string_at(
                        (current + 1), 1, {"C", "K", "Q"}
                    ) and not _string_at((current + 1), 2, {"CE", "CI"}):
                        current += 2
                    else:
                        current += 1
                    continue

            elif _get_at(current) == "D":
                if _string_at(current, 2, {"DG"}):
                    if _string_at((current + 2), 1, {"I", "E", "Y"}):
                        # e.g. 'edge'
                        primary, secondary = _metaph_add("J")
                        current += 3
                        continue
                    else:
                        # e.g. 'edgar'
                        primary, secondary = _metaph_add("TK")
                        current += 2
                        continue

                elif _string_at(current, 2, {"DT", "DD"}):
                    primary, secondary = _metaph_add("T")
                    current += 2
                    continue

                # else
                else:
                    primary, secondary = _metaph_add("T")
                    current += 1
                    continue

            elif _get_at(current) == "F":
                if _get_at(current + 1) == "F":
                    current += 2
                else:
                    current += 1
                primary, secondary = _metaph_add("F")
                continue

            elif _get_at(current) == "G":
                if _get_at(current + 1) == "H":
                    if (current > 0) and not _is_vowel(current - 1):
                        primary, secondary = _metaph_add("K")
                        current += 2
                        continue

                    # 'ghislane', ghiradelli
                    elif current == 0:
                        if _get_at(current + 2) == "I":
                            primary, secondary = _metaph_add("J")
                        else:
                            primary, secondary = _metaph_add("K")
                        current += 2
                        continue

                    # Parker's rule (with some further refinements) -
                    # e.g., 'hugh'
                    elif (
                        (
                            (current > 1)
                            and _string_at((current - 2), 1, {"B", "H", "D"})
                        )
                        # e.g., 'bough'
                        or (
                            (current > 2)
                            and _string_at((current - 3), 1, {"B", "H", "D"})
                        )
                        # e.g., 'broughton'
                        or ((current > 3) and _string_at((current - 4), 1, {"B", "H"}))
                    ):
                        current += 2
                        continue
                    else:
                        # e.g. 'laugh', 'McLaughlin', 'cough',
                        #      'gough', 'rough', 'tough'
                        if (
                            (current > 2)
                            and (_get_at(current - 1) == "U")
                            and (
                                _string_at((current - 3), 1, {"C", "G", "L", "R", "T"})
                            )
                        ):
                            primary, secondary = _metaph_add("F")
                        elif (current > 0) and _get_at(current - 1) != "I":
                            primary, secondary = _metaph_add("K")
                        current += 2
                        continue

                elif _get_at(current + 1) == "N":
                    if (current == 1) and _is_vowel(0) and not _slavo_germanic():
                        primary, secondary = _metaph_add("KN", "N")
                    # not e.g. 'cagney'
                    elif (
                        not _string_at((current + 2), 2, {"EY"})
                        and (_get_at(current + 1) != "Y")
                        and not _slavo_germanic()
                    ):
                        primary, secondary = _metaph_add("N", "KN")
                    else:
                        primary, secondary = _metaph_add("KN")
                    current += 2
                    continue

                # 'tagliaro'
                elif _string_at((current + 1), 2, {"LI"}) and not _slavo_germanic():
                    primary, secondary = _metaph_add("KL", "L")
                    current += 2
                    continue

                # -ges-, -gep-, -gel-, -gie- at beginning
                elif (current == 0) and (
                    (_get_at(current + 1) == "Y")
                    or _string_at(
                        (current + 1),
                        2,
                        {
                            "ES",
                            "EP",
                            "EB",
                            "EL",
                            "EY",
                            "IB",
                            "IL",
                            "IN",
                            "IE",
                            "EI",
                            "ER",
                        },
                    )
                ):
                    primary, secondary = _metaph_add("K", "J")
                    current += 2
                    continue

                #  -ger-,  -gy-
                elif (
                    (
                        _string_at((current + 1), 2, {"ER"})
                        or (_get_at(current + 1) == "Y")
                    )
                    and not _string_at(0, 6, {"DANGER", "RANGER", "MANGER"})
                    and not _string_at((current - 1), 1, {"E", "I"})
                    and not _string_at((current - 1), 3, {"RGY", "OGY"})
                ):
                    primary, secondary = _metaph_add("K", "J")
                    current += 2
                    continue

                #  italian e.g, 'biaggi'
                elif _string_at((current + 1), 1, {"E", "I", "Y"}) or _string_at(
                    (current - 1), 4, {"AGGI", "OGGI"}
                ):
                    # obvious germanic
                    if (
                        _string_at(0, 4, {"VAN ", "VON "}) or _string_at(0, 3, {"SCH"})
                    ) or _string_at((current + 1), 2, {"ET"}):
                        primary, secondary = _metaph_add("K")
                    elif _string_at((current + 1), 4, {"IER "}):
                        primary, secondary = _metaph_add("J")
                    else:
                        primary, secondary = _metaph_add("J", "K")
                    current += 2
                    continue

                else:
                    if _get_at(current + 1) == "G":
                        current += 2
                    else:
                        current += 1
                    primary, secondary = _metaph_add("K")
                    continue

            elif _get_at(current) == "H":
                # only keep if first & before vowel or btw. 2 vowels
                if ((current == 0) or _is_vowel(current - 1)) and _is_vowel(
                    current + 1
                ):
                    primary, secondary = _metaph_add("H")
                    current += 2
                else:  # also takes care of 'HH'
                    current += 1
                continue

            elif _get_at(current) == "J":
                # obvious spanish, 'jose', 'san jacinto'
                if _string_at(current, 4, {"JOSE"}) or _string_at(0, 4, {"SAN "}):
                    if ((current == 0) and (_get_at(current + 4) == " ")) or _string_at(
                        0, 4, {"SAN "}
                    ):
                        primary, secondary = _metaph_add("H")
                    else:
                        primary, secondary = _metaph_add("J", "H")
                    current += 1
                    continue

                elif (current == 0) and not _string_at(current, 4, {"JOSE"}):
                    # Yankelovich/Jankelowicz
                    primary, secondary = _metaph_add("J", "A")
                # Spanish pron. of e.g. 'bajador'
                elif (
                    _is_vowel(current - 1)
                    and not _slavo_germanic()
                    and ((_get_at(current + 1) == "A") or (_get_at(current + 1) == "O"))
                ):
                    primary, secondary = _metaph_add("J", "H")
                elif current == last:
                    primary, secondary = _metaph_add("J", " ")
                elif not _string_at(
                    (current + 1), 1, {"L", "T", "K", "S", "N", "M", "B", "Z"}
                ) and not _string_at((current - 1), 1, {"S", "K", "L"}):
                    primary, secondary = _metaph_add("J")

                if _get_at(current + 1) == "J":  # it could happen!
                    current += 2
                else:
                    current += 1
                continue

            elif _get_at(current) == "K":
                if _get_at(current + 1) == "K":
                    current += 2
                else:
                    current += 1
                primary, secondary = _metaph_add("K")
                continue

            elif _get_at(current) == "L":
                if _get_at(current + 1) == "L":
                    # Spanish e.g. 'cabrillo', 'gallegos'
                    if (
                        (current == (length - 3))
                        and _string_at((current - 1), 4, {"ILLO", "ILLA", "ALLE"})
                    ) or (
                        (
                            _string_at((last - 1), 2, {"AS", "OS"})
                            or _string_at(last, 1, {"A", "O"})
                        )
                        and _string_at((current - 1), 4, {"ALLE"})
                    ):
                        primary, secondary = _metaph_add("L", " ")
                        current += 2
                        continue
                    current += 2
                else:
                    current += 1
                primary, secondary = _metaph_add("L")
                continue

            elif _get_at(current) == "M":
                if (
                    (
                        _string_at((current - 1), 3, {"UMB"})
                        and (
                            ((current + 1) == last)
                            or _string_at((current + 2), 2, {"ER"})
                        )
                    )
                    # 'dumb', 'thumb'
                    or (_get_at(current + 1) == "M")
                ):
                    current += 2
                else:
                    current += 1
                primary, secondary = _metaph_add("M")
                continue

            elif _get_at(current) == "N":
                if _get_at(current + 1) == "N":
                    current += 2
                else:
                    current += 1
                primary, secondary = _metaph_add("N")
                continue

            elif _get_at(current) == "Ñ":
                current += 1
                primary, secondary = _metaph_add("N")
                continue

            elif _get_at(current) == "P":
                if _get_at(current + 1) == "H":
                    primary, secondary = _metaph_add("F")
                    current += 2
                    continue

                # also account for "campbell", "raspberry"
                elif _string_at((current + 1), 1, {"P", "B"}):
                    current += 2
                else:
                    current += 1
                primary, secondary = _metaph_add("P")
                continue

            elif _get_at(current) == "Q":
                if _get_at(current + 1) == "Q":
                    current += 2
                else:
                    current += 1
                primary, secondary = _metaph_add("K")
                continue

            elif _get_at(current) == "R":
                # french e.g. 'rogier', but exclude 'hochmeier'
                if (
                    (current == last)
                    and not _slavo_germanic()
                    and _string_at((current - 2), 2, {"IE"})
                    and not _string_at((current - 4), 2, {"ME", "MA"})
                ):
                    primary, secondary = _metaph_add("", "R")
                else:
                    primary, secondary = _metaph_add("R")

                if _get_at(current + 1) == "R":
                    current += 2
                else:
                    current += 1
                continue

            elif _get_at(current) == "S":
                # special cases 'island', 'isle', 'carlisle', 'carlysle'
                if _string_at((current - 1), 3, {"ISL", "YSL"}):
                    current += 1
                    continue

                # special case 'sugar-'
                elif (current == 0) and _string_at(current, 5, {"SUGAR"}):
                    primary, secondary = _metaph_add("X", "S")
                    current += 1
                    continue

                elif _string_at(current, 2, {"SH"}):
                    # Germanic
                    if _string_at((current + 1), 4, {"HEIM", "HOEK", "HOLM", "HOLZ"}):
                        primary, secondary = _metaph_add("S")
                    else:
                        primary, secondary = _metaph_add("X")
                    current += 2
                    continue

                # Italian & Armenian
                elif _string_at(current, 3, {"SIO", "SIA"}) or _string_at(
                    current, 4, {"SIAN"}
                ):
                    if not _slavo_germanic():
                        primary, secondary = _metaph_add("S", "X")
                    else:
                        primary, secondary = _metaph_add("S")
                    current += 3
                    continue

                # German & anglicisations, e.g. 'smith' match 'schmidt',
                #                               'snider' match 'schneider'
                # also, -sz- in Slavic language although in Hungarian it is
                #       pronounced 's'
                elif (
                    (current == 0)
                    and _string_at((current + 1), 1, {"M", "N", "L", "W"})
                ) or _string_at((current + 1), 1, {"Z"}):
                    primary, secondary = _metaph_add("S", "X")
                    if _string_at((current + 1), 1, {"Z"}):
                        current += 2
                    else:
                        current += 1
                    continue

                elif _string_at(current, 2, {"SC"}):
                    # Schlesinger's rule
                    if _get_at(current + 2) == "H":
                        # dutch origin, e.g. 'school', 'schooner'
                        if _string_at(
                            (current + 3),
                            2,
                            {"OO", "ER", "EN", "UY", "ED", "EM"},
                        ):
                            # 'schermerhorn', 'schenker'
                            if _string_at((current + 3), 2, {"ER", "EN"}):
                                primary, secondary = _metaph_add("X", "SK")
                            else:
                                primary, secondary = _metaph_add("SK")
                            current += 3
                            continue
                        else:
                            if (
                                (current == 0)
                                and not _is_vowel(3)
                                and (_get_at(3) != "W")
                            ):
                                primary, secondary = _metaph_add("X", "S")
                            else:
                                primary, secondary = _metaph_add("X")
                            current += 3
                            continue

                    elif _string_at((current + 2), 1, {"I", "E", "Y"}):
                        primary, secondary = _metaph_add("S")
                        current += 3
                        continue

                    # else
                    else:
                        primary, secondary = _metaph_add("SK")
                        current += 3
                        continue

                else:
                    # french e.g. 'resnais', 'artois'
                    if (current == last) and _string_at((current - 2), 2, {"AI", "OI"}):
                        primary, secondary = _metaph_add("", "S")
                    else:
                        primary, secondary = _metaph_add("S")

                    if _string_at((current + 1), 1, {"S", "Z"}):
                        current += 2
                    else:
                        current += 1
                    continue

            elif _get_at(current) == "T":
                if _string_at(current, 4, {"TION"}):
                    primary, secondary = _metaph_add("X")
                    current += 3
                    continue

                elif _string_at(current, 3, {"TIA", "TCH"}):
                    primary, secondary = _metaph_add("X")
                    current += 3
                    continue

                elif _string_at(current, 2, {"TH"}) or _string_at(current, 3, {"TTH"}):
                    # special case 'thomas', 'thames' or germanic
                    if (
                        _string_at((current + 2), 2, {"OM", "AM"})
                        or _string_at(0, 4, {"VAN ", "VON "})
                        or _string_at(0, 3, {"SCH"})
                    ):
                        primary, secondary = _metaph_add("T")
                    else:
                        primary, secondary = _metaph_add("0", "T")
                    current += 2
                    continue

                elif _string_at((current + 1), 1, {"T", "D"}):
                    current += 2
                else:
                    current += 1
                primary, secondary = _metaph_add("T")
                continue

            elif _get_at(current) == "V":
                if _get_at(current + 1) == "V":
                    current += 2
                else:
                    current += 1
                primary, secondary = _metaph_add("F")
                continue

            elif _get_at(current) == "W":
                # can also be in middle of word
                if _string_at(current, 2, {"WR"}):
                    primary, secondary = _metaph_add("R")
                    current += 2
                    continue
                elif (current == 0) and (
                    _is_vowel(current + 1) or _string_at(current, 2, {"WH"})
                ):
                    # Wasserman should match Vasserman
                    if _is_vowel(current + 1):
                        primary, secondary = _metaph_add("A", "F")
                    else:
                        # need Uomo to match Womo
                        primary, secondary = _metaph_add("A")

                # Arnow should match Arnoff
                if (
                    ((current == last) and _is_vowel(current - 1))
                    or _string_at(
                        (current - 1), 5, {"EWSKI", "EWSKY", "OWSKI", "OWSKY"}
                    )
                    or _string_at(0, 3, {"SCH"})
                ):
                    primary, secondary = _metaph_add("", "F")
                    current += 1
                    continue
                # Polish e.g. 'filipowicz'
                elif _string_at(current, 4, {"WICZ", "WITZ"}):
                    primary, secondary = _metaph_add("TS", "FX")
                    current += 4
                    continue
                # else skip it
                else:
                    current += 1
                    continue

            elif _get_at(current) == "X":
                # French e.g. breaux
                if not (
                    (current == last)
                    and (
                        _string_at((current - 3), 3, {"IAU", "EAU"})
                        or _string_at((current - 2), 2, {"AU", "OU"})
                    )
                ):
                    primary, secondary = _metaph_add("KS")

                if _string_at((current + 1), 1, {"C", "X"}):
                    current += 2
                else:
                    current += 1
                continue

            elif _get_at(current) == "Z":
                # Chinese Pinyin e.g. 'zhao'
                if _get_at(current + 1) == "H":
                    primary, secondary = _metaph_add("J")
                    current += 2
                    continue
                elif _string_at((current + 1), 2, {"ZO", "ZI", "ZA"}) or (
                    _slavo_germanic()
                    and ((current > 0) and _get_at(current - 1) != "T")
                ):
                    primary, secondary = _metaph_add("S", "TS")
                else:
                    primary, secondary = _metaph_add("S")

                if _get_at(current + 1) == "Z":
                    current += 2
                else:
                    current += 1
                continue

            else:
                current += 1

        if self._max_length > 0:
            primary = primary[: self._max_length]
            secondary = secondary[: self._max_length]
        if primary == secondary:
            secondary = ""

        return ",".join((primary, secondary))


if __name__ == "__main__":
    import doctest

    doctest.testmod()
